import numpy
import h5py
import os
try:
    from hyperion.dust import IsotropicDust
    from hyperion.model import Model as HypModel
    from hyperion.model import ModelOutput
    from hyperion.model.helpers import find_last_iteration
    from hyperion.util.otf_hdf5 import _on_the_fly_hdf5
    import h5py
except:
    print("WARNING: Hyperion versions <= 0.9.10 will not work with astropy >= 4.0 because they depend on astropy.extern.six. Continuing without Hyperion.")
from .. import radmc3d
from .Grid import Grid
from ..imaging import Image, UnstructuredImage, imtovis
from ..interferometry import Visibilities
from ..spectroscopy import Spectrum
from ..constants.astronomy import AU, M_sun, R_sun, L_sun, Jy, arcsec, pc
from ..constants.physics import c
import time

class Model:
    r"""
    A class that can be used to set up a generic model with arbitrary coordinates and densities.

    Base Attributes:
        :attr:`grid` (Grid):
            The Grid object that contains information about the 3D spatial distribution of the material in the system.
        :attr:`images` (dict):
            A dictionary containing the radiative transfer model images that have been generated for the model.
        :attr:`spectra` (dict):
            A dictionary containing the radiative transfer model spectra that have been generated for the model.
        :attr:`visibilities` (dict):
            A dictionary containing the radiative transfer model visibilities that have been generated for the model.
    """


    def __init__(self):
        self.grid = Grid()
        self.images = {}
        self.spectra = {}
        self.visibilities = {}

    def set_camera_wavelength(self, lam):
        r"""
        Set the wavelengths that the camera should use for generating images, visibilities, and spectra.

        Args:
            :attr:`lam` (numpy.ndarray `n`):
                The array of wavelengths, in microns.
        """
        self.camera_wavelength = lam

    def run_thermal(self, nphot=1e6, code="radmc3d", **keywords):
        r"""
        Run the radiative equilibrium calculation for the model.

        Args:
            :attr:`nphot` (int, optional):
                The number of photons to use for the calculation. Default: `1e6`
            :attr:`code` (str, optional):
                Which underlying radiative transfer modeling code to use for the radiative equilibrium calculation. `radmc3d` or `hyperion`. Default: `radmc3d`
            :attr:`kwargs` (optional):
                Can be used to pass arguments to either 
                :code:`Model.run_thermal_hyperion` or 
                :code:`Model.run_thermal_radmc3d`, depending on the value of 
                :code:`code`.
        """
        if (code == "radmc3d"):
            self.run_thermal_radmc3d(nphot=nphot, **keywords)
        else:
            self.run_thermal_hyperion(nphot=nphot, **keywords)

    def run_thermal_hyperion(self, nphot=1e6, mrw=False, pda=False, \
            niterations=20, percentile=99., absolute=2.0, relative=1.02, \
            max_interactions=1e8, mpi=False, nprocesses=None, \
            sublimation_temperature=None, verbose=True, timeout=3600, \
            increase_photons_until_convergence=False):
        """
        Run the radiative equilibrium calculation using the Hyperion radiative
        transfer code. As a result, the `Model.grid.temperature` list will be
        populated with the temperatures calculated.

        Args:
            :attr:`nphot` (int, optional):
                The number of photons to use for the calculation. Default: `1e6`
            :attr:`mrw` (bool, optional):
                If using `hyperion`, whether or not to use the Modified Random 
                Walk Algorithm. Default: `False`
            :attr:`pda` (bool, optional):
                If using `hyperion`, whether or not to use the Partial Diffusion
                Approximation algorithm. Default: `False`
            :attr:`niterations` (int, optional):
                The maximum number of iterations to perform, if convergence is
                not reached, before giving up. Default: `20`
            :attr:`percentile` (`float`, optional):
                Which percentile of cells to use when calculating the 
                convergence criteria. Default: `99.`
            :attr:`absolute` (`float`, optional):
                Maximum absolute difference of the ratio between cells for 
                convergence to be reached. Default: `2.0`
            :attr:`relative` (`float`, optional):
                Relative difference between cells for convergence to be reached.
                Default: `1.02`
            :attr:`max_interactions` (`int`, optional):
                Maximum number of interactions a photon can have before it is 
                killed. Default: `1e8`
            :attr:`mpi` (bool, optional):
                If using `hyperion`, whether or not to run the model in 
                parallel with MPI. Default: `False`
            :attr:`nprocesses` (bool, optional):
                If `mpi=True`, the number of MPI threads to use. Default: `None`
            :attr:`sublimation_temperature` (`float`, optional):
                If you would like to sublimate dust above a certain temperature,
                set that here. Default: `None`
            :attr:`verbose` (`bool`, False):
                Should output be printed to the screen, or hidden. 
                Default: `False`
        """

        d = []
        for i in range(len(self.grid.dust)):
            d.append(IsotropicDust( \
                    self.grid.dust[i].nu[::-1].astype(numpy.float64), \
                    self.grid.dust[i].albedo[::-1].astype(numpy.float64), \
                    self.grid.dust[i].kext[::-1].astype(numpy.float64)))

            if sublimation_temperature != None:
                d[-1].set_sublimation_temperature('fast', \
                        temperature=sublimation_temperature)

        m = HypModel()
        if (self.grid.coordsystem == "cartesian"):
            m.set_cartesian_grid(self.grid.w1*AU, self.grid.w2*AU, \
                    self.grid.w3*AU)
        elif (self.grid.coordsystem == "cylindrical"):
            m.set_cylindrical_polar_grid(self.grid.w1*AU, self.grid.w3*AU, \
                    self.grid.w2)
        elif (self.grid.coordsystem == "spherical"):
            m.set_spherical_polar_grid(self.grid.w1*AU, self.grid.w2, \
                    self.grid.w3)

        for i in range(len(self.grid.density)):
            if (self.grid.coordsystem == "cartesian"):
                m.add_density_grid(numpy.transpose(self.grid.density[i], \
                        axes=(2,1,0)), d[i])
            if (self.grid.coordsystem == "cylindrical"):
                m.add_density_grid(numpy.transpose(self.grid.density[i], \
                        axes=(1,2,0)), d[i])
            if (self.grid.coordsystem == "spherical"):
                m.add_density_grid(numpy.transpose(self.grid.density[i], \
                        axes=(2,1,0)), d[i])

        sources = []
        for i in range(len(self.grid.stars)):
            sources.append(m.add_point_source())
            sources[i].luminosity = self.grid.stars[i].luminosity * L_sun
            #sources[i].radius = self.grid.stars[i].radius * R_sun
            sources[i].temperature = self.grid.stars[i].temperature

        m.set_mrw(mrw)
        m.set_pda(pda)
        m.set_max_interactions(max_interactions)
        m.set_n_initial_iterations(niterations)
        m.conf.output.output_density = 'last'
        m.set_convergence(True, percentile=percentile, absolute=absolute, \
                relative=relative)

        converged = False
        while not converged:
            m.set_n_photons(initial=nphot, imaging=0)
            m.write("temp.rtin")

            if verbose:
                m.run("temp.rtout", mpi=mpi, n_processes=nprocesses, \
                        overwrite=True, timeout=timeout)
            else:
                m.run("temp.rtout", mpi=mpi, n_processes=nprocesses, \
                        overwrite=True, logfile="temp.log", timeout=timeout)

            f = h5py.File("temp.rtout","r")
            if verbose:
                print("Ran ", find_last_iteration(f), "iterations")
            if increase_photons_until_convergence and \
                    find_last_iteration(f) == niterations:
                if verbose:
                    print("Convergence not reached, increasing nphot by a factor of 10 and trying again.")
                nphot *= 10
                m.use_quantities("temp.rtout", quantities=['specific_energy'])
                os.system("rm temp.rtin temp.rtout temp.log")
            else:
                converged = True
            f.close()

        n = ModelOutput("temp.rtout")

        grid = n.get_quantities()

        self.grid.temperature = []
        temperature = grid.quantities['temperature']
        density = grid.quantities['density']
        for i in range(len(temperature)):
            if (self.grid.coordsystem == "cartesian"):
                self.grid.temperature.append(numpy.transpose(temperature[i], \
                        axes=(2,1,0)))
                self.grid.density[i] = numpy.transpose(density[i], \
                        axes=(2,1,0))
            if (self.grid.coordsystem == "cylindrical"):
                self.grid.temperature.append(numpy.transpose(temperature[i], \
                        axes=(2,0,1)))
                self.grid.density[i] = numpy.transpose(density[i], \
                        axes=(2,0,1))
            if (self.grid.coordsystem == "spherical"):
                self.grid.temperature.append(numpy.transpose(temperature[i], \
                        axes=(2,1,0)))
                self.grid.density[i] = numpy.transpose(density[i], \
                        axes=(2,1,0))

        os.system("rm temp.rtin temp.rtout temp.log")

    def run_thermal_radmc3d(self, nphot=1e6, verbose=True, timelimit=7200, \
            nice=None, **keywords):
        """
        Run the radiative equilibrium calculation using the RADMC-3D radiative
        transfer code. As a result, the `Model.grid.temperature` list will be
        populated with the temperatures calculated.

        Args:
            :attr:`nphot` (int, optional):
                The number of photons to use for the calculation. Default: `1e6`
            :attr:`verbose` (`bool`, False):
                Should output be printed to the screen, or hidden. 
                Default: `False`
            :attr:`**keywords` (optional):
                This can be used to pass any options to RADMC-3D. For a list of
                all possibilities for the thermal code, check the RADMC-3D
                documentation. Two commonly used ones are described below.
            :attr:`modified_random_walk` (bool, optional):
                If using `radmc3d`, whether or not to use the Modified Random 
                Walk Algorithm. Default: `False`
            :attr:`setthreads` (int, optional):
                The number of OpenMP threads to use. Default: `1`
        """

        self.write_radmc3d(nphot_therm=nphot, **keywords)

        radmc3d.run.thermal(verbose=verbose, timelimit=timelimit, nice=nice)

        self.grid.temperature = radmc3d.read.dust_temperature()
        for i in range(len(self.grid.temperature)):
            n1, n2, n3 = self.grid.density[i].shape
            self.grid.temperature[i] = numpy.transpose( \
                    self.grid.temperature[i].reshape((n3,n2,n1)), \
                    axes=(2,1,0))

        os.system("rm *.out *.inp *.dat")

    def run_scattering(self, nphot=1e6, code="radmc3d", **keywords):
        r"""
        Run a scattering phase function calculation for the model.

        Args:
            :attr:`nphot` (int, optional):
                The number of photons to use for the calculation. Default: `1e6`
            :attr:`code` (str, optional):
                Which underlying radiative transfer modeling code to use for 
                the radiative equilibrium calculation. `radmc3d` or `hyperion`. 
                Default: `radmc3d`
            :attr:`kwargs` (optional):
                Can be used to pass arguments to either 
                :code:`Model.run_scattering_hyperion` or 
                :code:`Model.run_scattering_radmc3d`, depending on the value of 
                :code:`code`.
        """
        if (code == "radmc3d"):
            self.run_scattering_radmc3d(nphot=nphot, **keywords)
        else:
            print("Scattering phase function cannot be calculated in Hyperion!")

    def run_scattering_radmc3d(self, nphot=1e6, verbose=True, nice=None, \
            loadlambda=None, **keywords):
        """
        Run a scattering phase function calculation using the RADMC-3D radiative
        transfer code. As a result, the `Model.grid.scattering_phase` list will 
        be populated with the scattering phase functions calculated.

        Args:
            :attr:`nphot` (int, optional):
                The number of photons to use for the calculation. Default: `1e6`
            :attr:`verbose` (`bool`, False):
                Should output be printed to the screen, or hidden. 
                Default: `False`
            :attr:`loadlambda` (`bool`, optional):
                If `True`, use the :code:`Model.camera_wavelength` array as the
                list of wavelengths for the scattering phase function 
                calculation. If `None`, use the :code:`Model.grid.lam array`. 
                Default: `None`
            :attr:`**keywords` (optional):
                This can be used to pass any options to RADMC-3D. For a list of
                all possibilities for the scattering phase function code, check 
                the RADMC-3D documentation. Two commonly used ones are 
                described below.
            :attr:`setthreads` (int, optional):
                The number of OpenMP threads to use. Default: `1`
        """

        self.write_radmc3d(nphot_scat=nphot, **keywords)

        radmc3d.run.scattering(verbose=verbose, nice=nice, \
                loadlambda=loadlambda)

        self.grid.scattering_phase_freq, self.grid.scattering_phase = \
                radmc3d.read.scattering_phase()
        for i in range(len(self.grid.scattering_phase)):
            n1, n2, n3 = self.grid.density[0].shape
            self.grid.scattering_phase[i] = numpy.transpose( \
                    self.grid.scattering_phase[i].reshape((n3,n2,n1)), \
                    axes=(2,1,0))

        self.grid.scattering_phase = numpy.array(self.grid.scattering_phase)

        os.system("rm *.out *.inp *.dat")

    def run_image(self, name=None, nphot=1e6, code="radmc3d", **keywords):
        r"""
        Run an image of the model.

        Args:
            :attr:`name` (`str`, optional):
                The name of the image, to use as a key in the 
                :code:`Model.images` dictionary. Default: `None`
            :attr:`nphot` (int, optional):
                The number of photons to use for the calculation. Default: `1e6`
            :attr:`code` (str, optional):
                Which underlying radiative transfer modeling code to use for 
                the radiative equilibrium calculation. `radmc3d` or `hyperion`. 
                Default: `radmc3d`
            :attr:`kwargs` (optional):
                Can be used to pass arguments to either 
                :code:`Model.run_scattering_hyperion` or 
                :code:`Model.run_scattering_radmc3d`, depending on the value of 
                :code:`code`.
        """

        if (code == "radmc3d"):
            self.run_image_radmc3d(name=name, nphot=nphot, **keywords)
        else:
            self.run_image_hyperion(name=name, nphot=nphot, **keywords)

    def run_image_hyperion(self, nphot=1e6, mrw=False, pda=False, \
            niterations=20, percentile=99., absolute=2.0, relative=1.02, \
            max_interactions=1e8, mpi=False, nprocesses=None, \
            sublimation_temperature=None, verbose=True, incl=45, pa=45, \
            dpc=1, lam=1300., track_origin='basic', nphot_imaging=1e6, \
            name=None, npix=256, pixelsize=1.0):
        """
        Run the radiative equilibrium calculation using the Hyperion radiative
        transfer code. As a result, the `Model.grid.temperature` list will be
        populated with the temperatures calculated.

        Args:
            :attr:`name` (`str`, optional):
                The name of the spectrum, to use as a key in the 
                :code:`Model.spectra` dictionary. Default: `None`
            :attr:`nphot` (int, optional):
                The number of photons to use for the calculation. Default: `1e6`
            :attr:`mrw` (bool, optional):
                If using `hyperion`, whether or not to use the Modified Random 
                Walk Algorithm. Default: `False`
            :attr:`pda` (bool, optional):
                If using `hyperion`, whether or not to use the Partial Diffusion
                Approximation algorithm. Default: `False`
            :attr:`niterations` (int, optional):
                The maximum number of iterations to perform, if convergence is
                not reached, before giving up. Default: `20`
            :attr:`percentile` (`float`, optional):
                Which percentile of cells to use when calculating the 
                convergence criteria. Default: `99.`
            :attr:`absolute` (`float`, optional):
                Maximum absolute difference of the ratio between cells for 
                convergence to be reached. Default: `2.0`
            :attr:`relative` (`float`, optional):
                Relative difference between cells for convergence to be reached.
                Default: `1.02`
            :attr:`max_interactions` (`int`, optional):
                Maximum number of interactions a photon can have before it is 
                killed. Default: `1e8`
            :attr:`mpi` (bool, optional):
                If using `hyperion`, whether or not to run the model in 
                parallel with MPI. Default: `False`
            :attr:`nprocesses` (bool, optional):
                If `mpi=True`, the number of MPI threads to use. Default: `None`
            :attr:`sublimation_temperature` (`float`, optional):
                If you would like to sublimate dust above a certain temperature,
                set that here. Default: `None`
            :attr:`verbose` (`bool`, False):
                Should output be printed to the screen, or hidden. 
                Default: `False`
            :attr:`incl` (`float`, optional):
                The inclination of the model to use for the image. Default: `0.`
            :attr:`pa` (`float`, optional):
                The position angle to use for the image. Default: `0.`
            :attr:`dpc` (`float`, optional):
                The distance to the model in units of parsecs. Default: `1.`
            :attr:`lam` (`float`, optional):
                The wavelength, in microns, to make the image at. Default: 
                `1300.`
            :attr:`nphot_imaging` (`float`, optional):
                The number of photons to use in making the image. Default: `1e6`
            :attr:`npix` (`int`, optional):
                The number of pixels (squared) that the image should have. 
                Default: `256`
            :attr:`pixelsize` (`float`, optional):
                The size of the pixels, in arcseconds. Default: `1.`
        """

        d = []
        for i in range(len(self.grid.dust)):
            d.append(IsotropicDust( \
                    self.grid.dust[i].nu[::-1].astype(numpy.float64), \
                    self.grid.dust[i].albedo[::-1].astype(numpy.float64), \
                    self.grid.dust[i].kext[::-1].astype(numpy.float64)))

            if sublimation_temperature != None:
                d[-1].set_sublimation_temperature('fast', \
                        temperature=sublimation_temperature)

        m = HypModel()
        if (self.grid.coordsystem == "cartesian"):
            m.set_cartesian_grid(self.grid.w1*AU, self.grid.w2*AU, \
                    self.grid.w3*AU)
        elif (self.grid.coordsystem == "cylindrical"):
            m.set_cylindrical_polar_grid(self.grid.w1*AU, self.grid.w3*AU, \
                    self.grid.w2)
        elif (self.grid.coordsystem == "spherical"):
            m.set_spherical_polar_grid(self.grid.w1*AU, self.grid.w2, \
                    self.grid.w3)

        for i in range(len(self.grid.density)):
            if (self.grid.coordsystem == "cartesian"):
                m.add_density_grid(numpy.transpose(self.grid.density[i], \
                        axes=(2,1,0)), d[i])
            if (self.grid.coordsystem == "cylindrical"):
                m.add_density_grid(numpy.transpose(self.grid.density[i], \
                        axes=(1,2,0)), d[i])
            if (self.grid.coordsystem == "spherical"):
                m.add_density_grid(numpy.transpose(self.grid.density[i], \
                        axes=(2,1,0)), d[i])

        sources = []
        for i in range(len(self.grid.stars)):
            sources.append(m.add_spherical_source())
            sources[i].luminosity = self.grid.stars[i].luminosity * L_sun
            sources[i].radius = self.grid.stars[i].radius * R_sun
            sources[i].temperature = self.grid.stars[i].temperature

        m.set_raytracing(True)
        
        if npix%2 == 0:
            zoomau = [-pixelsize*dpc*AU * (npix+1)/2, \
                    pixelsize*dpc*AU * (npix-1)/2, \
                    -pixelsize*dpc*AU * (npix+1)/2, \
                    pixelsize*dpc*AU * (npix-1)/2]
        else:
            zoomau = [-pixelsize*dpc*AU * npix/2, \
                    pixelsize*dpc*AU * npix/2, \
                    -pixelsize*dpc*AU * npix/2, \
                    pixelsize*dpc*AU * npix/2]

        m.set_monochromatic(True, wavelengths=[lam])
        image = m.add_peeled_images(sed=False, image=True)
        image.set_viewing_angles([incl], [pa])
        image.set_track_origin(track_origin)
        image.set_image_size(npix, npix)
        image.set_image_limits(*tuple(zoomau))

        m.set_mrw(mrw)
        m.set_pda(pda)
        m.set_max_interactions(max_interactions)
        m.set_n_initial_iterations(niterations)
        m.set_n_photons(initial=nphot, raytracing_sources=1e4, \
                raytracing_dust=nphot_imaging, imaging_sources=nphot_imaging, \
                imaging_dust=nphot_imaging)
        m.conf.output.output_density = 'last'
        m.set_convergence(True, percentile=percentile, absolute=absolute, \
                relative=relative)

        m.write("temp.rtin")

        if verbose:
            m.run("temp.rtout", mpi=mpi, n_processes=nprocesses, overwrite=True)
        else:
            m.run("temp.rtout", mpi=mpi, n_processes=nprocesses, \
                    overwrite=True, logfile="temp.log")

        n = ModelOutput("temp.rtout")

        image = n.get_image(inclination=0, distance=dpc*pc, units='Jy')

        self.images[name] = Image(image.val.reshape(image.val.shape+(1,)), \
                wave=image.wav*1.0e-4)

        if track_origin == "detailed":
            for component in ["source_emit","source_scat","dust_emit", \
                    "dust_scat"]:
                image = n.get_image(inclination=0, distance=dpc*pc, \
                        units='Jy', component=component)

                self.images[name+"-"+component] = Image(image.val.reshape(\
                        image.val.shape+(1,)), wave=image.wav*1.0e-4)

        os.system("rm temp.rtin temp.rtout temp.log")


    def run_image_radmc3d(self, name=None, nphot=1e6, npix=256, pixelsize=1.0, \
            lam="1300", loadlambda=False, imolspec=None, iline=None,  \
            widthkms=None, vkms=None, linenlam=None, doppcatch=False, \
            incl=0, pa=0, phi=0, dpc=1, verbose=True, nice=None, \
            unstructured=False, nostar=False, **keywords):
        self.write_radmc3d(nphot_scat=nphot, **keywords)

        if npix%2 == 0:
            zoomau = [-pixelsize*dpc * (npix+1)/2, pixelsize*dpc * (npix-1)/2, \
                    -pixelsize*dpc * (npix+1)/2, pixelsize*dpc * (npix-1)/2]
        else:
            zoomau = [-pixelsize*dpc * npix/2, pixelsize*dpc * npix/2, \
                    -pixelsize*dpc * npix/2, pixelsize*dpc * npix/2]

        radmc3d.run.image(npix=npix, zoomau=zoomau, lam=lam, \
                loadlambda=loadlambda, imolspec=imolspec, iline=iline, \
                widthkms=widthkms, vkms=vkms, linenlam=linenlam, \
                doppcatch=doppcatch, incl=incl, posang=pa, phi=phi, \
                verbose=verbose, nice=nice, circ=unstructured, \
                nostar=nostar)

        if unstructured:
            if 'writeimage_unformatted' in keywords:
                image, r, phi, lam = radmc3d.read.circimage(\
                        binary=keywords["writeimage_unformatted"])
            else:
                image, r, phi, lam = radmc3d.read.circimage()

            r, phi = numpy.meshgrid(r, phi)
            
            r = r / (dpc*pc) / arcsec

            x = -r*numpy.cos(phi)
            y = r*numpy.sin(phi)

            x = numpy.concatenate(([x[0,0]], x[:,1:].reshape((x[:,1:].size,))))
            y = numpy.concatenate(([y[0,0]], y[:,1:].reshape((y[:,1:].size,))))
            image = numpy.concatenate((image[0:1,0,:,0], image[:,1:,:,0].\
                    reshape((image[:,1:,0,0].size,image.shape[2]))))

            image = image / Jy

            self.images[name] = UnstructuredImage(image, x=x, y=y, \
                    wave=lam*1.0e-4)
        else:
            if 'writeimage_unformatted' in keywords:
                image, x, y, lam = radmc3d.read.image(\
                        binary=keywords["writeimage_unformatted"])
            else:
                image, x, y, lam = radmc3d.read.image()

            image = image / Jy * ((x[1] - x[0]) / (dpc * pc)) * \
                    ((y[1] - y[0]) / (dpc * pc))

            x = (x - x[int(npix/2)]) * pixelsize / (x[1] - x[0])
            y = (y - y[int(npix/2)]) * pixelsize / (y[1] - y[0])

            self.images[name] = Image(image, x=x, y=y, wave=lam*1.0e-4)

        os.system("rm *.out *.inp *.dat")
        if 'writeimage_unformatted' in keywords:
            if keywords['writeimage_unformatted']:
                os.system("rm *.bout")

    def run_sed(self, name=None, nphot=1e6, code="radmc3d", **keywords):
        r"""
        Run a spectrum of the model.

        Args:
            :attr:`name` (`str`, optional):
                The name of the spectrum, to use as a key in the 
                :code:`Model.spectra` dictionary. Default: `None`
            :attr:`nphot` (int, optional):
                The number of photons to use for the calculation. Default: `1e6`
            :attr:`code` (str, optional):
                Which underlying radiative transfer modeling code to use for 
                the radiative equilibrium calculation. `radmc3d` or `hyperion`. 
                Default: `radmc3d`
            :attr:`kwargs` (optional):
                Can be used to pass arguments to either 
                :code:`Model.run_sed_hyperion` or 
                :code:`Model.run_sed_radmc3d`, depending on the value of 
                :code:`code`.
        """

        if (code == "radmc3d"):
            self.run_sed_radmc3d(name=name, nphot=nphot, **keywords)
        else:
            self.run_sed_hyperion(name=name, nphot=nphot, **keywords)

    def run_sed_hyperion(self, name=None, nphot=1e6):
        return

    def run_sed_radmc3d(self, name=None, nphot=1e6, incl=0, pa=0, \
            phi=0, dpc=1, loadlambda=False, verbose=True, nice=None, \
            **keywords):
        r"""
        Run a spectrum of the model with RADMC-3D. As a result, the 
        :code:`Model.spectra` dictionary will have a 
        :code:`pdspy.spectroscopy.Spectrum` added with key `name`.

        Args:
            :attr:`name` (`str`, optional):
                The name of the spectrum, to use as a key in the 
                :code:`Model.spectra` dictionary. Default: `None`
            :attr:`nphot` (int, optional):
                The number of photons to use for the calculation. Default: `1e6`
            :attr:`incl` (`float`, optional):
                The inclination of the model to use for the image. Default: `0.`
            :attr:`loadlambda` (`bool`, optional):
                If `True`, use the :code:`Model.camera_wavelength` array as the
                list of wavelengths for the scattering phase function 
                calculation. If `False`, use the :code:`Model.grid.lam array`. 
                Default: `False`
            :attr:`code` (str, optional):
                Which underlying radiative transfer modeling code to use for 
                the radiative equilibrium calculation. `radmc3d` or `hyperion`. 
                Default: `radmc3d`
            :attr:`kwargs` (optional):
                Can be used to pass arguments to RADMC-3D.
                This can be used to pass any options to RADMC-3D. For a list of
                all possibilities for spectra, check the RADMC-3D documentation.
        """

        self.write_radmc3d(nphot_spec=nphot, **keywords)

        radmc3d.run.sed(incl=incl, posang=pa, phi=phi, noline=True, \
                loadlambda=loadlambda, verbose=verbose, nice=nice)

        flux, lam = radmc3d.read.spectrum()

        flux = flux / Jy * (1. / dpc)**2

        self.spectra[name] = Spectrum(wave=lam, flux=flux)

        os.system("rm *.out *.inp *.dat")

    def run_visibilities(self, name=None, nphot=1e6, code="radmc3d", \
            **keywords):
        r"""
        Run visibilities for the model.

        Args:
            :attr:`name` (`str`, optional):
                The name of the visibilities, to use as a key in the 
                :code:`Model.visibilities` dictionary. Default: `None`
            :attr:`nphot` (int, optional):
                The number of photons to use for the calculation. Default: `1e6`
            :attr:`code` (str, optional):
                Which underlying radiative transfer modeling code to use for 
                the radiative equilibrium calculation. `radmc3d` or `hyperion`. 
                Default: `radmc3d`
            :attr:`kwargs` (optional):
                Can be used to pass arguments to either 
                :code:`Model.run_scattering_hyperion` or 
                :code:`Model.run_scattering_radmc3d`, depending on the value of 
                :code:`code`.
        """

        if (code == "radmc3d"):
            self.run_visibilities_radmc3d(name=name, nphot=nphot, **keywords)
        else:
            self.run_visibilities_hyperion(name=name, nphot=nphot, **keywords)

    def run_visibilities_hyperion(self, name=None, nphot=1e6):
        return

    def run_visibilities_radmc3d(self, name=None, nphot=1e6, npix=256, \
            pixelsize=1.0, lam="1300", loadlambda=False, imolspec=None, \
            iline=None,  widthkms=None, vkms=None, linenlam=None, \
            doppcatch=False, incl=0, pa=0, phi=0, dpc=1, verbose=True, \
            nice=None, nostar=False, **keywords):
        self.write_radmc3d(nphot_scat=nphot, **keywords)

        if npix%2 == 0:
            zoomau = [-pixelsize*dpc * (npix+1)/2, pixelsize*dpc * (npix-1)/2, \
                    -pixelsize*dpc * (npix+1)/2, pixelsize*dpc * (npix-1)/2]
        else:
            zoomau = [-pixelsize*dpc * npix/2, pixelsize*dpc * npix/2, \
                    -pixelsize*dpc * npix/2, pixelsize*dpc * npix/2]

        radmc3d.run.image(npix=npix, zoomau=zoomau, lam=lam, \
                loadlambda=loadlambda, imolspec=imolspec, iline=iline, \
                widthkms=widthkms, vkms=vkms, linenlam=linenlam, \
                doppcatch=doppcatch, incl=incl, posang=pa, phi=phi, \
                verbose=verbose, nostar=nostar, nice=nice)

        if 'writeimage_unformatted' in keywords:
            image, x, y, lam = radmc3d.read.image(\
                    binary=keywords["writeimage_unformatted"])
        else:
            image, x, y, lam = radmc3d.read.image()

        image = image / Jy * ((x[1] - x[0]) / (dpc * pc)) * \
                ((y[1] - y[0]) / (dpc * pc))

        x = x * pixelsize / (x[1] - x[0])
        y = y * pixelsize / (y[1] - y[0])

        im = Image(image, x=x, y=y, wave=lam*1.0e-4)

        self.visibilities[name] = imtovis(im)

        os.system("rm *.out *.inp *.dat")
        if 'writeimage_unformatted' in keywords:
            if keywords['writeimage_unformatted']:
                os.system("rm *.bout")

    def write_radmc3d(self, **keywords):
        radmc3d.write.control(**keywords)

        mstar = []
        rstar = []
        xstar = []
        ystar = []
        zstar = []
        tstar = []

        for i in range(len(self.grid.stars)):
            mstar.append(self.grid.stars[i].mass*M_sun)
            rstar.append(self.grid.stars[i].radius*R_sun)
            xstar.append(self.grid.stars[i].x*AU)
            ystar.append(self.grid.stars[i].y*AU)
            zstar.append(self.grid.stars[i].z*AU)
            tstar.append(self.grid.stars[i].temperature)

        radmc3d.write.stars(rstar, mstar, self.grid.lam, xstar, ystar, zstar, \
                tstar=tstar)

        radmc3d.write.wavelength_micron(self.grid.lam)
        if hasattr(self, "camera_wavelength"):
            radmc3d.write.camera_wavelength_micron(self.camera_wavelength)

        if (self.grid.coordsystem == "cartesian"):
            radmc3d.write.amr_grid(self.grid.w1*AU, self.grid.w2*AU, \
                    self.grid.w3*AU, coordsystem=self.grid.coordsystem)
        elif(self.grid.coordsystem == "cylindrical"):
            radmc3d.write.amr_grid(self.grid.w1*AU, self.grid.w2, \
                    self.grid.w3*AU, coordsystem=self.grid.coordsystem)
        elif(self.grid.coordsystem == "spherical"):
            radmc3d.write.amr_grid(self.grid.w1*AU, self.grid.w2, \
                    self.grid.w3, coordsystem=self.grid.coordsystem)

        radmc3d.write.dust_density(self.grid.density)
        if len(self.grid.temperature) > 0:
            density = numpy.array(self.grid.density)
            temperature = numpy.array(self.grid.temperature)

            density[density == 0] = 1.0e-30

            temperature = (density * temperature).sum(axis=0) / \
                    density.sum(axis=0)
            temperature = [temperature for i in range(len(self.grid.density))]

            radmc3d.write.dust_temperature(temperature)

        dustopac = []
        for i in range(len(self.grid.dust)):
            dustopac.append("dustkappa_{0:d}.inp".format(i))
            radmc3d.write.dustkappa("{0:d}".format(i), \
                    self.grid.dust[i].lam*1.0e4, self.grid.dust[i].kabs, \
                    ksca=self.grid.dust[i].ksca)

        radmc3d.write.dustopac(dustopac)

        if len(self.grid.gas) > 0:
            gas = []
            inpstyle = []
            colpartners = []
            for i in range(len(self.grid.gas)):
                gas.append("{0:d}".format(i))
                inpstyle.append("leiden")
                colpartners.append([])
                radmc3d.write.molecule(self.grid.gas[i], gas[i])
                radmc3d.write.numberdens(self.grid.number_density[i], gas[i])

            radmc3d.write.line(gas, inpstyle, colpartners)

            number_density = numpy.array(self.grid.number_density)
            velocity = numpy.array(self.grid.velocity)
            vx = velocity[:,0,:,:,:]
            vy = velocity[:,1,:,:,:]
            vz = velocity[:,2,:,:,:]
            velocity = numpy.zeros(self.grid.velocity[0].shape)

            nx, ny, nz = self.grid.number_density[0].shape

            number_density[number_density == 0] = 1.0e-50

            velocity[0,:,:,:] = (number_density * vx).sum(axis=0) / \
                    number_density.sum(axis=0)
            velocity[1,:,:,:] = (number_density * vy).sum(axis=0) / \
                    number_density.sum(axis=0)
            velocity[2,:,:,:] = (number_density * vz).sum(axis=0) / \
                    number_density.sum(axis=0)

            radmc3d.write.gas_velocity(velocity)

            if len(self.grid.gas_temperature) > 0:
                gas_temperature = numpy.array(self.grid.gas_temperature)
                gas_temperature = (number_density * gas_temperature).\
                        sum(axis=0) / number_density.sum(axis=0)

                radmc3d.write.gas_temperature(gas_temperature)

            if len(self.grid.microturbulence) > 0:
                microturbulence = numpy.array(self.grid.microturbulence)
                microturbulence = (number_density * microturbulence).\
                        sum(axis=0) / number_density.sum(axis=0)

                radmc3d.write.microturbulence(microturbulence)

    def read(self, filename=None, usefile=None):
        r"""
        Read a model in from an HDF5 model file.

        Args:
            :attr:`filename` (`str`, optional):
                The filename of the file storing the model to read in. 
                Default: `None`
            :attr:`usefile` (`h5py.File`, optional):
                If :code:`filename = None`, then use this keyword to provide an
                instance of an :code:`h5py.File` that has already be opened 
                that the model can be read from.
        """

        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        if ('Grid' in f):
            self.grid = Grid()
            self.grid.read(usefile=f['Grid'])

        if ('Images' in f):
            for image in f['Images']:
                try:
                    self.images[image] = Image()
                    self.images[image].read(usefile=f['Images'][image])
                except:
                    self.images[image] = UnstructuredImage()
                    self.images[image].read(usefile=f['Images'][image])

        if ('Spectra' in f):
            for spectrum in f['Spectra']:
                self.spectra[spectrum] = Spectrum()
                self.spectra[spectrum].read(usefile=f['Spectra'][spectrum])

        if ('Visibilities' in f):
            for visibility in f['Visibilities']:
                self.visibilities[visibility] = Visibilities()
                self.visibilities[visibility].read( \
                        usefile=f['Visibilities'][visibility])

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        r"""
        Write a model to an HDF5 model file.

        Args:
            :attr:`filename` (`str`, optional):
                The filename of the file that will be written out with the 
                model. Default: `None`
            :attr:`usefile` (`h5py.File`, optional):
                If :code:`filename = None`, then use this keyword to provide an
                instance of an :code:`h5py.File` that has already be opened 
                that the model can be written to.
        """

        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        grid = f.create_group("Grid")
        if hasattr(self, 'grid'):
            self.grid.write(usefile=grid)

        spectra = f.create_group("Spectra")
        for spectrum in self.spectra:
            spectra.create_group(spectrum)
            self.spectra[spectrum].write(usefile=spectra[spectrum])

        images = f.create_group("Images")
        for image in self.images:
            images.create_group(image)
            self.images[image].write(usefile=images[image])

        visibilities = f.create_group("Visibilities")
        for visibility in self.visibilities:
            visibilities.create_group(visibility)
            self.visibilities[visibility].write( \
                    usefile=visibilities[visibility])

        if (usefile == None):
            f.close()



from ..constants.physics import c
from .libinterferometry import Visibilities
import astropy.table
import casatools
import numpy

def readms(filename, spw='all', tolerance=0.01, time_tolerance=0., \
        datacolumn="corrected", corr=["I"]):
    """Testing

    :param filename: The name of the MS file that you would like to read in.
    :param spw: The list of spectral windows to read in.
    """

    # Load the MS file.

    ms = casatools.ms()

    ms.open(filename)

    # If we want all of the spw, then get them all.

    if spw == 'all':
        tb = casatools.table()

        tb.open(filename)
        spw = tb.getcol('DATA_DESC_ID')
        tb.close()

        spw = list(numpy.unique(spw))

    # Loop through all of the DATA_DESC_ID values and collect the relevant data.

    if datacolumn == "corrected":
        prefix = "corrected_"
    else:
        prefix = ""

    i = 0                                                               
    ms.reset()                                                          
    data = []
    for i in spw:
        ms.selectinit(datadescid=i)
        ms.selectpolarization(corr)
        data.append(ms.getdata(items=["u","v",prefix+"real",prefix+\
                "imaginary","weight","flag","axis_info","uvdist",\
                "antenna1","antenna2","time"]))
        ms.reset()

    # We are done with the data now, so close it.

    ms.close()

    # Test how many of the spectral windows are identical to each other.

    matching_spw = [0]

    for i in range(1,len(data)):
        matched = False

        for j in range(i):
            if i == j:
                continue

            # If they don't have the same number of frequencies, they can't be
            # the same spectral window.

            if data[i]["axis_info"]["freq_axis"]["chan_freq"].shape != \
                    data[j]["axis_info"]['freq_axis']['chan_freq'].shape:
                continue

            # Test how different they are and compare with some threshold.

            diff = (data[i]["axis_info"]['freq_axis']['chan_freq'] - \
                    data[j]["axis_info"]['freq_axis']['chan_freq']) / \
                    data[j]["axis_info"]["freq_axis"]['resolution']

            if numpy.abs(diff).max() < tolerance:
                matching_spw.append(j)
                matched = True
                break

        # If we didn't find a match, add as a new spectral window.

        if not matched:
            matching_spw.append(i)

    # Loop through all of the unique spectral windows that we found.

    for i in numpy.unique(matching_spw):
        new_freq = data[i]["axis_info"]["freq_axis"]["chan_freq"]

        # Collect all of the DATA_DESC_ID's that match that spectral window.

        new_u, new_v, new_real, new_imag, new_weights, new_flags, \
                new_uvdist, new_antenna1, new_antenna2, new_time = [], [], [], \
                [], [], [], [], [], [], []

        for j in range(len(data)):
            if matching_spw[j] == i:
                new_u.append(data[j]["u"])
                new_v.append(data[j]["v"])
                new_uvdist.append(data[j]["uvdist"])
                new_real.append(data[j][prefix+"real"])
                new_imag.append(data[j][prefix+"imaginary"])
                new_weights.append(data[j]["weight"])
                new_flags.append(data[j]["flag"])
                new_antenna1.append(data[j]["antenna1"])
                new_antenna2.append(data[j]["antenna2"])
                new_time.append(data[j]["time"])

        # Concatenate them all along the uvdist axis.

        new_u = numpy.concatenate(new_u)
        new_v = numpy.concatenate(new_v)
        new_uvdist = numpy.concatenate(new_uvdist)
        new_real = numpy.concatenate(new_real, axis=2)
        new_imag = numpy.concatenate(new_imag, axis=2)
        new_weights = numpy.concatenate(new_weights, axis=1)
        new_flags = numpy.concatenate(new_flags, axis=2)
        new_antenna1 = numpy.concatenate(new_antenna1)
        new_antenna2 = numpy.concatenate(new_antenna2)
        new_time = numpy.concatenate(new_time)

        # Adjust the weights for flags and make the right shape.
        
        new_weights = ((new_flags == False) * new_weights.reshape((\
                new_weights.shape[0],1,new_weights.shape[1])))

        # Average out the two different polarizations.

        new_real = (new_real * new_weights).sum(axis=0)
        new_imag = (new_imag * new_weights).sum(axis=0)
        new_weights = new_weights.sum(axis=0)

        new_real[new_weights != 0] /= new_weights[new_weights != 0]
        new_imag[new_weights != 0] /= new_weights[new_weights != 0]

        # Transpose to make the right shape.

        new_real = numpy.transpose(new_real)
        new_imag = numpy.transpose(new_imag)
        new_weights = numpy.transpose(new_weights)

        # Trim the autocorrelation data.

        good = new_uvdist != 0

        new_u = new_u[good]
        new_v = new_v[good]
        new_uvdist = new_uvdist[good]
        new_real = new_real[good,:]
        new_imag = new_imag[good,:]
        new_weights = new_weights[good,:]
        new_antenna1 = new_antenna1[good]
        new_antenna2 = new_antenna2[good]
        new_time = new_time[good]

        # Now do some fancy stuff to merge the multiple spectral windows. Since 
        # it is possible that even for the same observation, there'll be a 
        # different number of UV points because of flagging, we need to match 
        # the uv points that exist for both, but add zeros when that point 
        # doesn't exist for one or the other.

        if i == 0:
            freq = new_freq

            # Make the first column uvdist.

            table = astropy.table.Table([new_antenna1, \
                    new_antenna2, new_time, new_u, new_v, new_real, new_imag, \
                    new_weights, new_uvdist], names=["antenna1","antenna2",\
                    "time"]+[string+"_"+str(i) for string in ['u','v','real',\
                    'imag','weights','uvdist']])
        else:
            freq = numpy.concatenate((freq, new_freq))

            # Make the first column uvdist.

            new_table = astropy.table.Table([new_antenna1, \
                    new_antenna2, new_time, new_u, new_v, new_real, new_imag, \
                    new_weights, new_uvdist], names=["antenna1","antenna2",\
                    "time"]+[string+"_"+str(i) for string in ['u','v','real',\
                    'imag','weights','uvdist']])

            # Do the fancy merging.

            if time_tolerance == 0:
                table = astropy.table.join(table, new_table, join_type='outer',\
                        keys=["antenna1","antenna2","time"])
            else:
                table = astropy.table.join(table, new_table, join_type='outer',\
                        keys=["antenna1","antenna2","time"], \
                        join_funcs={'time':astropy.table.\
                        join_distance(time_tolerance)})

                for key in ["time"]:
                    table[key] = numpy.concatenate([\
                            table[key+'_1'].data[:,numpy.newaxis], \
                            table[key+'_2'].data[:,numpy.newaxis]], axis=1)

                    wt = numpy.where(table[key].data == 0, 0, 1)

                    table[key] = (table[key].data*wt).sum(axis=1)/\
                            wt.sum(axis=1)
                
                    table.remove_columns([key+'_1',key+'_2',key+'_id'])

    # Make sure masked colums are filled appropriately.

    for colname in table.colnames:
        table[colname].fill_value = 0.
    table = table.filled()

    # Get the info out of the table.

    u = numpy.concatenate([table['u_{0:d}'.format(i)].data[:,numpy.newaxis] \
            for i in numpy.unique(matching_spw)], axis=1)
    v = numpy.concatenate([table['v_{0:d}'.format(i)].data[:,numpy.newaxis] \
            for i in numpy.unique(matching_spw)], axis=1)

    freq = freq[:,0]

    real = numpy.concatenate([table['real_{0:d}'.format(i)].data for i in \
            numpy.unique(matching_spw)], axis=1)
    imag = numpy.concatenate([table['imag_{0:d}'.format(i)].data for i in \
            numpy.unique(matching_spw)], axis=1)
    weights = numpy.concatenate([table['weights_{0:d}'.format(i)].data for i \
            in numpy.unique(matching_spw)], axis=1)

    # Trim down u and v.

    uwt = numpy.where(numpy.logical_and(u == 0, v == 0), 0, 1)
    u = (u*uwt).sum(axis=1) / uwt.sum(axis=1)
    v = (v*uwt).sum(axis=1) / uwt.sum(axis=1)

    # Include the complex conjugate.

    scale = 100 * freq.mean() / c

    u = numpy.concatenate((u, -u))*scale
    v = numpy.concatenate((v, -v))*scale
    real = numpy.concatenate((real, real))
    imag = -1*numpy.concatenate((imag, -imag))
    weights = numpy.concatenate((weights, weights))
    
    return Visibilities(u, v, freq, real, imag, weights)



import numpy
import scipy.special
from ..constants.astronomy import arcsec
from ..constants.physics import c
from .libinterferometry import Visibilities

def model(u, v, params, return_type="complex", funct="gauss", freq=230., \
        primary_beam=None):
    
    if type(params) == list:
        params = numpy.array(params)
    elif type(params) == numpy.ndarray:
        pass

    if type(funct) == str:
        funct = numpy.array([funct])
    elif type(funct) == list:
        funct = numpy.array(funct)
    elif type(funct) == numpy.ndarray:
        pass

    nparams = numpy.zeros(funct.size)
    for i in range(funct.size):
        if funct[i] == "point":
            nparams[i] = 3
        elif funct[i] == "gauss":
            nparams[i] = 6
        elif funct[i] == "circle":
            nparams[i] = 6
        elif funct[i] == "ring":
            nparams[i] = 7
    
    model = 1j*numpy.zeros(u.size)
    for i in range(funct.size):
        index = 0
        for j in range(i):
            index += nparams[j]
        index = int(index)
        
        # Make a copy of the parameters so we don't mess it up.
        par = params.copy()

        # If we were suplied with a primary beam shape, then calculate the flux
        # correction factor.
        if primary_beam != None:
            r0 = (par[index+0]**2 + par[index+1]**2)**0.5

            if primary_beam == "ALMA":
                fwhm_pb = 1.13 * (c*1.0e-2 / freq) / 12 / arcsec
            elif primary_beam == "VLA":
                fwhm_pb = 60. * 45 / (freq / 1.0e9)
            else:
                fwhm_pb = 0.

            sigma_pb = fwhm_pb / (2. * numpy.sqrt(2 * numpy.log(2.)))
            pb = numpy.exp(-r0**2/(2*sigma_pb**2))

            par[int(index + nparams[i] - 1)] *= pb

        # Convert sizes into arcseconds.
        par[index+0] *= arcsec
        par[index+1] *= arcsec
        if (funct[i] == "gauss") or (funct[i] == "circle") or \
                (funct[i] == "ring"):
            par[index+2] *= arcsec
            #if (funct[i] == "gauss"):
            if (funct[i] == "gauss") or (funct[i] == "ring"):
                par[index+3] *= arcsec
        
        # Generate the model.
        if funct[i] == "point":
            model += point_model(u, v, par[index+0], par[index+1], par[index+2])
        elif funct[i] == "gauss":
            model += gaussian_model(u, v, par[index+0], par[index+1], \
                par[index+2], par[index+3], par[index+4], par[index+5])
        elif funct[i] == "circle":
            model += circle_model(u, v, par[index+0], par[index+1], \
                par[index+2], par[index+3], par[index+4], par[index+5])
        elif funct[i] == "ring":
            model += ring_model(u, v, par[index+0], par[index+1], \
                par[index+2], par[index+3], par[index+4], par[index+5], \
                par[index+6])

    real = model.real
    imag = model.imag
    
    if return_type == "real":
        return real
    elif return_type == "imag":
        return imag
    elif return_type == "complex":
        return (real + 1j*imag).reshape((real.size,1))
    elif return_type == "amp":
        return numpy.sqrt(real**2 + imag**2)
    elif return_type == "data":
        return Visibilities(u, v, numpy.array([freq]), \
                real.reshape((real.size,1)), imag.reshape((imag.size,1)), \
                numpy.ones((real.size,1)))
    elif return_type == "append":
        return numpy.concatenate((real,imag))

def point_model(u, v, xcenter, ycenter, flux):
    
    return flux * numpy.exp(-2*3.14159 * (0 + 1j*(u * xcenter +v * ycenter)))

def circle_model(u, v, xcenter, ycenter, radius, incline, theta, flux):
    
    urot = u * numpy.cos(theta) - v * numpy.sin(theta)
    vrot = u * numpy.sin(theta) + v * numpy.cos(theta)

    numpy.seterr(invalid="ignore")

    vis = flux * scipy.special.j1(2*numpy.pi * radius * \
            numpy.sqrt(urot**2+vrot**2*numpy.cos(incline)**2)) / \
            (numpy.pi * radius * numpy.sqrt(urot**2+vrot**2*\
            numpy.cos(incline)**2)) * numpy.exp(-2*numpy.pi * (0 + \
            1j*(u*xcenter+v*ycenter)))

    numpy.seterr(invalid="warn")

    vis[urot**2+vrot**2*numpy.cos(incline)**2 == 0] = flux

    return vis

def ring_model(u, v, xcenter, ycenter, inradius, outradius, incline, theta, \
        flux):

    if inradius > outradius:
        return numpy.repeat(numpy.inf, u.size)

    A = (inradius/outradius)**2

    return flux * (circle_model(u, v, xcenter, ycenter, outradius, incline, \
            theta, 1) - A*circle_model(u, v, xcenter, ycenter, inradius, \
            incline, theta, 1)) / (1 - A)

def gaussian_model(u, v, xcenter, ycenter, usigma, vsigma, theta, flux):

    #if vsigma > usigma:
    #    return numpy.repeat(numpy.inf, u.size)
    
    urot = u * numpy.cos(theta) - v * numpy.sin(theta)
    vrot = u * numpy.sin(theta) + v * numpy.cos(theta)
    
    return flux * numpy.exp(-2*numpy.pi**2 * (usigma**2 * urot**2 + \
            vsigma**2 * vrot**2)) * numpy.exp(-2*numpy.pi * \
            1j*(u*xcenter+v*ycenter))



import numpy

def freefree(nu, F_nu_ff, nu_turn, pl_turn):
    if type(nu) == numpy.ndarray:
        flux = numpy.where(nu < nu_turn, F_nu_ff * (nu / nu_turn)**pl_turn, \
                F_nu_ff * (nu / nu_turn)**-0.1)

        flux = numpy.where(nu > 1.0e13, 0., flux)

        return flux
    else:
        if nu < nu_turn:
            return F_nu_ff * (nu / nu_turn)**pl_turn
        elif nu > 1.0e13:
            return 0.
        else:
            return F_nu_ff * (nu / nu_turn)**-0.1



import numpy
import h5py
from ..constants.physics import G, m_p
from ..constants.astronomy import AU, M_sun
from ..constants.math import pi
from ..dust import Dust
from ..gas import Gas
from .SettledDisk import SettledDisk

class SettledPringleDisk(SettledDisk):
    def surface_density(self, r, normalize=True):
        # Get the disk parameters.

        rr = r * AU
        rin = self.rmin * AU
        rout = self.rmax * AU
        mass = self.mass * M_sun
        gamma = self.plrho - self.plh
        if self.gamma_taper != None:
            gamma_taper = self.gamma_taper
        else:
            gamma_taper = gamma

        # Set up the surface density.

        Sigma0 = (2-gamma)*mass/(2*pi*rout**2)*numpy.exp((rin/rout)**(2-gamma))

        Sigma = Sigma0 * (rr/rout)**(-gamma) * \
                numpy.exp(-(rr/rout)**(2-gamma_taper))

        Sigma[r <= rin/AU] = 0e0

        # In case of r == 0 (a singularity), get the value from slightly off 0.

        dr = rr[r > 0].min()
        Sigma[r == 0] = Sigma0 * (0.7*dr/rout)**(-gamma) * \
                numpy.exp(-(0.7*dr/rout)**(2-gamma))

        # Add gaps to the disk.

        for i in range(len(self.gap_rin)):
            if self.gaussian_gaps:
                gap_r = (self.gap_rin[i] + self.gap_rout[i])/2
                gap_w = self.gap_rout[i] - self.gap_rin[i]

                Sigma /= 1 + 1./self.gap_delta[i] * numpy.exp(-4*numpy.log(2.)*\
                        (r - gap_r)**2 / gap_w**2)
            else:
                Sigma[(r >= self.gap_rin[i]) & \
                        (r <= self.gap_rout[i])] *= self.gap_delta[i]
        
        ##### Normalize the surface density correctly.
        
        if normalize:
            r_high = numpy.logspace(numpy.log10(self.rmin), \
                    numpy.log10(self.rmax), 1000)
            Sigma_high = self.surface_density(r_high, normalize=False)

            scale = mass / (2*numpy.pi*trapz(r_high*AU*Sigma_high, r_high*AU))

            Sigma *= scale

        return Sigma

    def scale_height(self, r):
        return self.h0 * AU * (r / self.rmax)**self.plh

    def temperature(self, r, theta, phi):
        ##### Disk Parameters
        
        rin = self.rmin * AU
        rout = self.rmax * AU
        t0 = self.t0
        plt = self.plt

        ##### Set up the coordinates

        rt, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        rr = rt*numpy.sin(tt)
        zz = rt*numpy.cos(tt)

        ##### Make the dust density model for a protoplanetary disk.
        
        t = t0 * (rr / (1*AU))**(-plt)

        t[rr <= rin] = 0e0

        t[t > 10000] = 10000
        
        return t

    def temperature_1d(self, r):
        rin = self.rmin * AU
        rout = self.rmax * AU
        t0 = self.t0
        plt = self.plt

        T = t0 * r**(-plt)

        T[r <= rin/AU] = 0.0

        dr = r[r > 0].min()
        T[r == 0] = t0 * (0.7*dr)**(-plt)

        return T

    def gas_temperature(self, r, theta, phi):
        ##### Disk Parameters
        
        rin = self.rmin * AU
        rout = self.rmax * AU
        pltgas = self.pltgas
        tmid0 = self.tmid0
        tatm0 = self.tatm0
        zq0 = self.zq0
        delta = self.delta

        ##### Set up the coordinates

        rt, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        rr = rt*numpy.sin(tt)
        zz = rt*numpy.cos(tt)

        ##### Make the dust density model for a protoplanetary disk.
        
        zq = zq0 * (rt / rin)**1.3

        tmid = tmid0 * (rr / rin)**(-pltgas)
        tatm = tatm0 * (rr / rin)**(-pltgas)

        t = numpy.zeros(tatm.shape)
        t[zz >= zq] = tatm[zz >= zq]
        t[zz < zq] = tatm[zz < zq] + (tmid[zz < zq] - tatm[zz < zq]) * \
                (numpy.cos(numpy.pi * zz[zz < zq] / (2*zq[zz < zq])))**2*delta
        
        return t



#!/usr/bin/env python3

from pdspy.dust import *
import numpy

water_ice = Dust()
water_ice.set_optical_constants_from_henn("optical_constants/water_ice.txt")
water_ice.set_density(0.92)

organics = Dust()
organics.set_optical_constants_from_henn("optical_constants/amorphous_carbon_zubko1996_extrapolated.txt")
organics.set_density(2.24)
organics.calculate_optical_constants_on_wavelength_grid(water_ice.lam)

silicates = Dust()
silicates.set_optical_constants_from_draine("optical_constants/astronomical_silicates.txt")
silicates.set_density(3.3)
silicates.calculate_optical_constants_on_wavelength_grid(water_ice.lam)

species = [silicates,organics,water_ice]
mass_fraction = numpy.array([2.64e-3,3.53e-3,5.55e-3])
rho = numpy.array([silicates.rho,organics.rho,water_ice.rho])
abundances = (mass_fraction/rho)/(mass_fraction/rho).sum()
print(abundances)

dust = mix_dust(species, abundances, filling=0.6)

amin = 0.1e-4
amax = 1.000e-4
pl = 3.5

dust.calculate_size_distribution_opacity(amin, amax, pl, with_dhs=True, \
        coat_volume_fraction=0.0)

dust.write('ricci_1um.hdf5')



class Transform:
    def __init__(self, xmin, xmax, dx, fmt):
        self.xmin = xmin
        self.xmax = xmax
        self.dx = dx
        self.fmt = fmt

    def __call__(self, x, p):
        return self.fmt% ((x-(self.xmax-self.xmin+1)/2)*self.dx)



import numpy
import h5py
from .Model import Model
from .Disk import Disk
from .DartoisDisk import DartoisDisk
from .TwoLayerDisk import TwoLayerDisk
from .SettledDisk import SettledDisk
from .PringleDisk import PringleDisk
from .DartoisPringleDisk import DartoisPringleDisk
from .SettledPringleDisk import SettledPringleDisk
from .TwoLayerPringleDisk import TwoLayerPringleDisk
from .Envelope import Envelope
from .UlrichEnvelope import UlrichEnvelope
from .UlrichEnvelopeExtended import UlrichEnvelopeExtended
from .TaperedUlrichEnvelope import TaperedUlrichEnvelope
from .TaperedUlrichEnvelopeExtended import TaperedUlrichEnvelopeExtended
from .Star import Star
from ..constants.physics import h, c, G, m_p, k
from ..constants.astronomy import AU, M_sun, kms, R_sun, Jy, pc
from ..constants.math import pi
from ..misc import B_nu
from ..imaging import Image, imtovis

class YSOModel(Model):
    r"""
    A Model that specifically represents a young star, including a star, disk, and envelope.
    """

    def add_star(self, mass=0.5, luminosity=1, temperature=4000.):
        self.grid.add_star(Star(mass=mass, luminosity=luminosity, \
                temperature=temperature))

    def set_cartesian_grid(self, xmin, xmax, nx):
        x = numpy.linspace(xmin, xmax, nx)
        y = numpy.linspace(xmin, xmax, nx)
        z = numpy.linspace(xmin, xmax, nx)

        self.grid.set_cartesian_grid(x, y, z)

    def set_cylindrical_grid(self, rmin, rmax, nr, nz, nphi):
        r = numpy.linspace(rmin, rmax, nr)
        phi = numpy.linspace(0.0, 2*numpy.pi, nphi)
        z = numpy.linspace(0.,rmax, nz)

        self.grid.set_cylindrical_grid(r, phi, z)

    def set_spherical_grid(self, rmin, rmax, nr, ntheta, nphi, log=True, \
            code="radmc3d"):
        if log:
            r = numpy.logspace(numpy.log10(rmin), numpy.log10(rmax), nr)
        else:
            r = numpy.linspace(rmin, rmax, nr)
        if (code == "hyperion"):
            r = numpy.hstack([0.0,r])

        if (code == "radmc3d"):
            theta = numpy.linspace(0.0, numpy.pi/2, ntheta)
        elif (code == "hyperion"):
            theta = numpy.linspace(0.0, numpy.pi, ntheta)

        phi = numpy.linspace(0.0, 2*numpy.pi, nphi)

        self.grid.set_spherical_grid(r, theta, phi)

    def add_ambient_medium(self, dens=1.0e-24):
        for i in range(len(self.grid.density)):
            self.grid.density[i] += dens

    def add_disk(self, mass=1.0e-3, rmin=0.1, rmax=300, plrho=2.37, h0=0.1, \
            plh=58./45., dust=None,  t0=None, plt=None, gas=None, \
            abundance=None, tmid0=None, tatm0=None, zq0=None, pltgas=None, \
            delta=None, gap_rin=[], gap_rout=[], gap_delta=[], \
            gaussian_gaps=False, aturb=None):
        self.disk = Disk(mass=mass, rmin=rmin, rmax=rmax, plrho=plrho, h0=h0, \
                plh=plh, dust=dust, t0=t0, plt=plt, tmid0=tmid0, tatm0=tatm0, \
                zq0=zq0, pltgas=pltgas, delta=delta, gap_rin=gap_rin, \
                gap_rout=gap_rout, gap_delta=gap_delta, aturb=aturb, \
                gaussian_gaps=gaussian_gaps)

        if (dust != None):
            self.grid.add_density(self.disk.density(self.grid.r, \
                    self.grid.theta, self.grid.phi),dust)

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.disk.add_gas(gas[i], abundance[i])
                    self.grid.add_number_density(self.disk.number_density(\
                            self.grid.r, self.grid.theta, self.grid.phi, \
                            gas=i), gas[i])
                    self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.disk.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.disk.add_gas(gas, abundance)
                self.grid.add_number_density(self.disk.number_density( \
                        self.grid.r, self.grid.theta, self.grid.phi, \
                        gas=0), gas)
                self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.disk.microturbulence(\
                            self.grid.r, self.grid.theta, self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.disk.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_gas_temperature(self.disk.gas_temperature( \
                    self.grid.r, self.grid.theta, self.grid.phi))

    def add_dartois_disk(self, mass=1.0e-3, rmin=0.1, rmax=300, plrho=2.37, \
            h0=0.1, plh=58./45., dust=None,  t0=None, plt=None, gas=None, \
            abundance=None, freezout=0., tmid0=None, tatm0=None, zq0=None, \
            pltgas=None, delta=None, gap_rin=[], gap_rout=[], gap_delta=[], \
            gaussian_gaps=False, aturb=None):
        self.disk = DartoisDisk(mass=mass, rmin=rmin, rmax=rmax, plrho=plrho, \
                h0=h0, plh=plh, dust=dust, t0=t0, plt=plt, tmid0=tmid0, \
                tatm0=tatm0, zq0=zq0, pltgas=pltgas, delta=delta, \
                gap_rin=gap_rin, gap_rout=gap_rout, gap_delta=gap_delta, \
                aturb=aturb, gaussian_gaps=gaussian_gaps)

        if (dust != None):
            self.grid.add_density(self.disk.density(self.grid.r, \
                    self.grid.theta, self.grid.phi),dust)

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.disk.add_gas(gas[i], abundance[i], freezeout[i])
                    self.grid.add_number_density(self.disk.number_density(\
                            self.grid.w1, self.grid.w2, self.grid.w3, \
                            gas=i, mstar=self.grid.stars[0].mass), gas[i])
                    self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.disk.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.disk.add_gas(gas, abundance, freezeout)
                self.grid.add_number_density(self.disk.number_density( \
                        self.grid.w1, self.grid.w2, self.grid.w3, \
                        gas=0, mstar=self.grid.stars[0].mass), gas)
                self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.disk.microturbulence(\
                            self.grid.r, self.grid.theta, self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.disk.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_temperature(self.disk.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))

    def add_twolayer_disk(self, mass=1.0e-3, rmin=0.1, rmax=300, plrho=2.37, \
            h0=0.1, plh=58./45., dust=None,  t0=None, plt=None, gas=None, \
            abundance=None, tmid0=None, tatm0=None, zq0=None, pltgas=None, \
            delta=None, gap_rin=[], gap_rout=[], gap_delta=[], \
            gaussian_gaps=False, aturb=None, amin=1., amax=1000., fmax=0.8, \
            alpha_settle=1.0e-3):
        self.disk = TwoLayerDisk(mass=mass, rmin=rmin, rmax=rmax, plrho=plrho, \
                h0=h0, plh=plh, dust=dust, t0=t0, plt=plt, tmid0=tmid0, \
                tatm0=tatm0, zq0=zq0, pltgas=pltgas, delta=delta, \
                gap_rin=gap_rin, gap_rout=gap_rout, gap_delta=gap_delta, \
                aturb=aturb, gaussian_gaps=gaussian_gaps, amin=amin, amax=amax,\
                fmax=fmax, alpha_settle=alpha_settle)

        if (dust != None):
            a, rho = self.disk.density(self.grid.r, self.grid.theta, \
                    self.grid.phi)

            for i in range(len(a)):
                self.grid.add_density(rho[:,:,:,i], self.disk.dust(a[i]/1e4, \
                        pla))

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.disk.add_gas(gas[i], abundance[i])
                    self.grid.add_number_density(self.disk.number_density(\
                            self.grid.r, self.grid.theta, self.grid.phi, \
                            gas=i), gas[i])
                    self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.disk.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.disk.add_gas(gas, abundance)
                self.grid.add_number_density(self.disk.number_density( \
                        self.grid.r, self.grid.theta, self.grid.phi, \
                        gas=0), gas)
                self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.disk.microturbulence(\
                            self.grid.r, self.grid.theta, self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.disk.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_gas_temperature(self.disk.gas_temperature( \
                    self.grid.r, self.grid.theta, self.grid.phi))

    def add_settled_disk(self, mass=1.0e-3, rmin=0.1, rmax=300, plrho=2.37, \
            h0=0.1, plh=58./45., dust=None,  t0=None, plt=None, gas=None, \
            abundance=None, tmid0=None, tatm0=None, zq0=None, pltgas=None, \
            delta=None, gap_rin=[], gap_rout=[], gap_delta=[], \
            gaussian_gaps=False, aturb=None, amin=0.05, amax=1000., pla=3.5, \
            alpha_settle=1.0e-3, na=100):
        self.disk = SettledDisk(mass=mass, rmin=rmin, rmax=rmax, plrho=plrho, \
                h0=h0, plh=plh, dust=dust, t0=t0, plt=plt, tmid0=tmid0, \
                tatm0=tatm0, zq0=zq0, pltgas=pltgas, delta=delta, \
                gap_rin=gap_rin, gap_rout=gap_rout, gap_delta=gap_delta, \
                aturb=aturb, gaussian_gaps=gaussian_gaps, amin=amin, amax=amax,\
                pla=pla, alpha_settle=alpha_settle)

        if (dust != None):
            a, rho = self.disk.density(self.grid.r, self.grid.theta, \
                    self.grid.phi, na=na)

            for i in range(len(a)):
                self.grid.add_density(rho[:,:,:,i], self.disk.dust(a[i]/1e4, \
                        pla))

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.disk.add_gas(gas[i], abundance[i])
                    self.grid.add_number_density(self.disk.number_density(\
                            self.grid.r, self.grid.theta, self.grid.phi, \
                            gas=i), gas[i])
                    self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.disk.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.disk.add_gas(gas, abundance)
                self.grid.add_number_density(self.disk.number_density( \
                        self.grid.r, self.grid.theta, self.grid.phi, \
                        gas=0), gas)
                self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.disk.microturbulence(\
                            self.grid.r, self.grid.theta, self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.disk.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_gas_temperature(self.disk.gas_temperature( \
                    self.grid.r, self.grid.theta, self.grid.phi))

    def add_pringle_disk(self, mass=1.0e-3, rmin=0.1, rmax=300, plrho=2.37, \
            h0=0.1, plh=58./45., dust=None,  t0=None, plt=None, gas=None, \
            abundance=None, tmid0=None, tatm0=None, zq0=None, pltgas=None, \
            delta=None, gap_rin=[], gap_rout=[], gap_delta=[], \
            gaussian_gaps=False, aturb=None, gamma_taper=None):
        self.disk = PringleDisk(mass=mass, rmin=rmin, rmax=rmax, plrho=plrho, \
                h0=h0, plh=plh, dust=dust, t0=t0, plt=plt, tmid0=tmid0, \
                tatm0=tatm0, zq0=zq0, pltgas=pltgas, delta=delta, \
                gap_rin=gap_rin, gap_rout=gap_rout, gap_delta=gap_delta, \
                aturb=aturb, gaussian_gaps=gaussian_gaps, \
                gamma_taper=gamma_taper)

        if (dust != None):
            if self.grid.coordsystem == "spherical":
                self.grid.add_density(self.disk.density(self.grid.r, \
                        self.grid.theta, self.grid.phi),dust)
            elif self.grid.coordsystem == "cartesian":
                self.grid.add_density(self.disk.density(self.grid.x, \
                        self.grid.y, self.grid.z, \
                        coordsys=self.grid.coordsystem),dust)
            elif self.grid.coordsystem == "cylindrical":
                self.grid.add_density(self.disk.density(self.grid.rho, \
                        self.grid.phi, self.grid.z, \
                        coordsys=self.grid.coordsystem),dust)

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.disk.add_gas(gas[i], abundance[i])
                    self.grid.add_number_density(self.disk.number_density( \
                            self.grid.r, self.grid.theta, self.grid.phi, \
                            gas=i), gas[i])
                    self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.disk.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.disk.add_gas(gas, abundance)
                self.grid.add_number_density(self.disk.number_density( \
                        self.grid.r, self.grid.theta, self.grid.phi, \
                        gas=0), gas)
                self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.disk.microturbulence( \
                            self.grid.r, self.grid.theta, self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.disk.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_gas_temperature(self.disk.gas_temperature( \
                    self.grid.r, self.grid.theta, self.grid.phi))

    def add_dartois_pringle_disk(self, mass=1.0e-3, rmin=0.1, rmax=300, \
            plrho=2.37, h0=0.1, plh=58./45., dust=None,  t0=None, plt=None, \
            gas=None, abundance=None, freezeout=0., tmid0=None, tatm0=None, \
            zq0=None, pltgas=None, delta=None, gap_rin=[], gap_rout=[], \
            gap_delta=[], gaussian_gaps=False, aturb=None, gamma_taper=None):
        self.disk = DartoisPringleDisk(mass=mass, rmin=rmin, rmax=rmax, \
                plrho=plrho, h0=h0, plh=plh, dust=dust, t0=t0, plt=plt, \
                tmid0=tmid0, tatm0=tatm0, zq0=zq0, pltgas=pltgas, delta=delta, \
                gap_rin=gap_rin, gap_rout=gap_rout, gap_delta=gap_delta, \
                aturb=aturb, gaussian_gaps=gaussian_gaps, \
                gamma_taper=gamma_taper)

        if (dust != None):
            if self.grid.coordsystem == "spherical":
                self.grid.add_density(self.disk.density(self.grid.r, \
                        self.grid.theta, self.grid.phi),dust)
            elif self.grid.coordsystem == "cartesian":
                self.grid.add_density(self.disk.density(self.grid.x, \
                        self.grid.y, self.grid.z, \
                        coordsys=self.grid.coordsystem),dust)
            elif self.grid.coordsystem == "cylindrical":
                self.grid.add_density(self.disk.density(self.grid.rho, \
                        self.grid.phi, self.grid.z, \
                        coordsys=self.grid.coordsystem),dust)

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.disk.add_gas(gas[i], abundance[i], freezeout[i])
                    self.grid.add_number_density(self.disk.number_density( \
                            self.grid.w1, self.grid.w2, self.grid.w3, \
                            gas=i, mstar=self.grid.stars[0].mass), gas[i])
                    self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.disk.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.disk.add_gas(gas, abundance, freezeout)
                self.grid.add_number_density(self.disk.number_density( \
                        self.grid.w1, self.grid.w2, self.grid.w3, \
                        gas=0, mstar=self.grid.stars[0].mass), gas)
                self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.disk.microturbulence( \
                            self.grid.r, self.grid.theta, self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.disk.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_temperature(self.disk.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))

    def add_twolayer_pringle_disk(self, mass=1.0e-3, rmin=0.1, rmax=300, \
            plrho=2.37, h0=0.1, plh=58./45., dust=None,  t0=None, plt=None, \
            gas=None, abundance=None, tmid0=None, tatm0=None, zq0=None, \
            pltgas=None, delta=None, gap_rin=[], gap_rout=[], gap_delta=[], \
            gaussian_gaps=False, aturb=None, amin=1., amax=1000., fmax=0.8, \
            alpha_settle=1.0e-3):
        self.disk = TwoLayerPringleDisk(mass=mass, rmin=rmin, rmax=rmax, \
                plrho=plrho, h0=h0, plh=plh, dust=dust, t0=t0, plt=plt, \
                tmid0=tmid0, tatm0=tatm0, zq0=zq0, pltgas=pltgas, delta=delta, \
                gap_rin=gap_rin, gap_rout=gap_rout, gap_delta=gap_delta, \
                aturb=aturb, gaussian_gaps=gaussian_gaps, amin=amin, amax=amax,\
                fmax=fmax, alpha_settle=alpha_settle)

        if (dust != None):
            a, rho = self.disk.density(self.grid.r, self.grid.theta, \
                    self.grid.phi)

            for i in range(len(a)):
                self.grid.add_density(rho[:,:,:,i], self.disk.dust(a[i]/1e4, \
                        pla))

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.disk.add_gas(gas[i], abundance[i])
                    self.grid.add_number_density(self.disk.number_density(\
                            self.grid.r, self.grid.theta, self.grid.phi, \
                            gas=i), gas[i])
                    self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.disk.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.disk.add_gas(gas, abundance)
                self.grid.add_number_density(self.disk.number_density( \
                        self.grid.r, self.grid.theta, self.grid.phi, \
                        gas=0), gas)
                self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.disk.microturbulence(\
                            self.grid.r, self.grid.theta, self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.disk.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_gas_temperature(self.disk.gas_temperature( \
                    self.grid.r, self.grid.theta, self.grid.phi))

    def add_settled_pringle_disk(self, mass=1.0e-3, rmin=0.1, rmax=300, \
            plrho=2.37, h0=0.1, plh=58./45., dust=None,  t0=None, plt=None, \
            gas=None, abundance=None, tmid0=None, tatm0=None, zq0=None, \
            pltgas=None, delta=None, gap_rin=[], gap_rout=[], gap_delta=[], \
            gaussian_gaps=False, aturb=None, amin=0.05, amax=1000., pla=3.5, \
            alpha_settle=1.0e-3, na=100, gamma_taper=None):
        self.disk = SettledPringleDisk(mass=mass, rmin=rmin, rmax=rmax, \
                plrho=plrho, h0=h0, plh=plh, dust=dust, t0=t0, plt=plt, \
                tmid0=tmid0, tatm0=tatm0, zq0=zq0, pltgas=pltgas, delta=delta, \
                gap_rin=gap_rin, gap_rout=gap_rout, gap_delta=gap_delta, \
                aturb=aturb, gaussian_gaps=gaussian_gaps, amin=amin, amax=amax,\
                pla=pla, alpha_settle=alpha_settle, gamma_taper=None)

        if (dust != None):
            a, rho = self.disk.density(self.grid.r, self.grid.theta, \
                    self.grid.phi, na=na)

            for i in range(len(a)):
                self.grid.add_density(rho[:,:,:,i], self.disk.dust(a[i]/1e4, \
                        pla))

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.disk.add_gas(gas[i], abundance[i])
                    self.grid.add_number_density(self.disk.number_density(\
                            self.grid.r, self.grid.theta, self.grid.phi, \
                            gas=i), gas[i])
                    self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.disk.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.disk.add_gas(gas, abundance)
                self.grid.add_number_density(self.disk.number_density( \
                        self.grid.r, self.grid.theta, self.grid.phi, \
                        gas=0), gas)
                self.grid.add_velocity(self.disk.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.disk.microturbulence( \
                            self.grid.r, self.grid.theta, self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.disk.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_gas_temperature(self.disk.gas_temperature( \
                    self.grid.r, self.grid.theta, self.grid.phi))

    def add_envelope(self, mass=1.0e-3, rmin=0.1, rmax=1000, pl=1.5, \
            cavpl=1.0, cavrfact=0.2, t0=None, tpl=None, dust=None, gas=None, \
            abundance=None, tmid0=None, aturb=None):
        self.envelope = Envelope(mass=mass, rmin=rmin, rmax=rmax, \
                pl=pl, cavpl=cavpl, cavrfact=cavrfact, t0=t0, tpl=tpl, \
                dust=dust)

        if (dust != None):
            self.grid.add_density(self.envelope.density(self.grid.r, \
                    self.grid.theta, self.grid.phi),dust)

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.envelope.add_gas(gas[i], abundance[i])
                    self.grid.add_number_density(self.envelope.number_density( \
                            self.grid.r, self.grid.theta, self.grid.phi, \
                            gas=i), gas[i])
                    self.grid.add_velocity(self.envelope.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.disk.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.envelope.add_gas(gas, abundance)
                self.grid.add_number_density(self.envelope.number_density( \
                        self.grid.r, self.grid.theta, self.grid.phi, gas=0), \
                        gas)
                self.grid.add_velocity(self.envelope.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))

                if aturb != None:
                    self.grid.add_microturbulence(self.disk.microturbulence( \
                            self.grid.r, self.grid.theta, self.grid.phi))
        if t0 != None:
            self.grid.add_temperature(self.envelope.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_gas_temperature(self.envelope.gas_temperature( \
                    self.grid.r, self.grid.theta, self.grid.phi))

    def add_ulrich_envelope(self, mass=1.0e-3, rmin=0.1, rmax=1000, rcent=300, \
            cavpl=1.0, cavrfact=0.2, t0=None, tpl=None, dust=None, gas=None, \
            abundance=None, tmid0=None, rcent_ne_rdisk=False, aturb=None):
        if rcent_ne_rdisk:
            self.envelope = UlrichEnvelope(mass=mass, rmin=rmin, rmax=rmax, \
                    rcent=rcent, cavpl=cavpl, cavrfact=cavrfact, t0=t0, \
                    tpl=tpl, dust=dust, aturb=aturb)
        elif hasattr(self, 'disk'):
            self.envelope = UlrichEnvelope(mass=mass, rmin=rmin, rmax=rmax, \
                    rcent=self.disk.rmax, cavpl=cavpl, cavrfact=cavrfact, \
                    t0=t0, tpl=tpl, dust=dust, aturb=aturb)
        else:
            self.envelope = UlrichEnvelope(mass=mass, rmin=rmin, rmax=rmax, \
                    rcent=rcent, cavpl=cavpl, cavrfact=cavrfact, t0=t0, \
                    tpl=tpl, dust=dust, aturb=aturb)

        if (dust != None):
            self.grid.add_density(self.envelope.density(self.grid.r, \
                    self.grid.theta, self.grid.phi),dust)

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.envelope.add_gas(gas[i], abundance[i])
                    self.grid.add_number_density(self.envelope.number_density( \
                            self.grid.r, self.grid.theta, self.grid.phi, \
                            gas=i), gas[i])
                    self.grid.add_velocity(self.envelope.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.envelope.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.envelope.add_gas(gas, abundance)
                self.grid.add_number_density(self.envelope.number_density( \
                        self.grid.r, self.grid.theta, self.grid.phi, gas=0), \
                        gas)
                self.grid.add_velocity(self.envelope.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.envelope.\
                            microturbulence(self.grid.r, self.grid.theta, \
                            self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.envelope.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_gas_temperature(self.envelope.gas_temperature( \
                    self.grid.r, self.grid.theta, self.grid.phi))

    def add_ulrichextended_envelope(self, mass=1.0e-3, rmin=0.1, rmax=1000, \
            rcent=300, cavpl=1.0, cavrfact=0.2, theta_open=45., zoffset=1., \
            t0=None, tpl=None, dust=None, gas=None, abundance=None, tmid0=None,\
            rcent_ne_rdisk=False, aturb=None):
        if rcent_ne_rdisk:
            self.envelope = UlrichEnvelopeExtended(mass=mass, rmin=rmin, \
                    rmax=rmax, rcent=rcent, cavpl=cavpl, cavrfact=cavrfact, \
                    theta_open=theta_open, zoffset=zoffset, t0=t0, tpl=tpl, \
                    dust=dust, aturb=aturb)
        elif hasattr(self, 'disk'):
            self.envelope = UlrichEnvelopeExtended(mass=mass, rmin=rmin, \
                    rmax=rmax, rcent=self.disk.rmax, cavpl=cavpl, \
                    cavrfact=cavrfact, theta_open=theta_open, zoffset=zoffset, \
                    t0=t0, tpl=tpl, dust=dust, aturb=aturb)
        else:
            self.envelope = UlrichEnvelopeExtended(mass=mass, rmin=rmin, \
                    rmax=rmax, rcent=rcent, cavpl=cavpl, cavrfact=cavrfact, \
                    theta_open=theta_open, zoffset=zoffset, t0=t0, tpl=tpl, \
                    dust=dust, aturb=aturb)

        if (dust != None):
            self.grid.add_density(self.envelope.density(self.grid.r, \
                    self.grid.theta, self.grid.phi),dust)

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.envelope.add_gas(gas[i], abundance[i])
                    self.grid.add_number_density(self.envelope.number_density( \
                            self.grid.r, self.grid.theta, self.grid.phi, \
                            gas=i), gas[i])
                    self.grid.add_velocity(self.envelope.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.envelope.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.envelope.add_gas(gas, abundance)
                self.grid.add_number_density(self.envelope.number_density( \
                        self.grid.r, self.grid.theta, self.grid.phi, gas=0), \
                        gas)
                self.grid.add_velocity(self.envelope.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.envelope.\
                            microturbulence(self.grid.r, self.grid.theta, \
                            self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.envelope.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_gas_temperature(self.envelope.gas_temperature( \
                    self.grid.r, self.grid.theta, self.grid.phi))

    def add_tapered_ulrich_envelope(self, mass=1.0e-3, rmin=0.1, rmax=1000, \
            rcent=300, cavpl=1.0, cavrfact=0.2, gamma=1., t0=None, tpl=None, \
            dust=None, gas=None, abundance=None, tmid0=None, \
            rcent_ne_rdisk=False, aturb=None):
        if rcent_ne_rdisk:
            self.envelope = TaperedUlrichEnvelope(mass=mass, rmin=rmin, \
                    rmax=rmax, rcent=rcent, gamma=gamma, cavpl=cavpl, \
                    cavrfact=cavrfact, t0=t0, tpl=tpl, dust=dust, aturb=aturb)
        elif hasattr(self, 'disk'):
            self.envelope = TaperedUlrichEnvelope(mass=mass, rmin=rmin, \
                    rmax=rmax, rcent=self.disk.rmax, gamma=gamma, cavpl=cavpl, \
                    cavrfact=cavrfact, t0=t0, tpl=tpl, dust=dust, aturb=aturb)
        else:
            self.envelope = TaperedUlrichEnvelope(mass=mass, rmin=rmin, \
                    rmax=rmax, rcent=rcent, gamma=gamma, cavpl=cavpl, \
                    cavrfact=cavrfact, t0=t0, tpl=tpl, dust=dust, aturb=aturb)

        if (dust != None):
            self.grid.add_density(self.envelope.density(self.grid.r, \
                    self.grid.theta, self.grid.phi),dust)

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.envelope.add_gas(gas[i], abundance[i])
                    self.grid.add_number_density(self.envelope.number_density( \
                            self.grid.r, self.grid.theta, self.grid.phi, \
                            gas=i), gas[i])
                    self.grid.add_velocity(self.envelope.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.envelope.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.envelope.add_gas(gas, abundance)
                self.grid.add_number_density(self.envelope.number_density( \
                        self.grid.r, self.grid.theta, self.grid.phi, gas=0), \
                        gas)
                self.grid.add_velocity(self.envelope.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.envelope.\
                            microturbulence(self.grid.r, self.grid.theta, \
                            self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.envelope.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_gas_temperature(self.envelope.gas_temperature( \
                    self.grid.r, self.grid.theta, self.grid.phi))

    def add_tapered_ulrichextended_envelope(self, mass=1.0e-3, rmin=0.1, \
            rmax=1000, rcent=300, cavpl=1.0, cavrfact=0.2, gamma=1., \
            theta_open=45., zoffset=1., t0=None, tpl=None, dust=None, gas=None,\
            abundance=None, tmid0=None, rcent_ne_rdisk=False, aturb=None):
        if rcent_ne_rdisk:
            self.envelope = TaperedUlrichEnvelopeExtended(mass=mass, rmin=rmin,\
                    rmax=rmax, rcent=rcent, gamma=gamma, cavpl=cavpl, \
                    cavrfact=cavrfact, theta_open=theta_open, zoffset=zoffset, \
                    t0=t0, tpl=tpl, dust=dust, aturb=aturb)
        elif hasattr(self, 'disk'):
            self.envelope = TaperedUlrichEnvelopeExtended(mass=mass, rmin=rmin,\
                    rmax=rmax, rcent=self.disk.rmax, gamma=gamma, cavpl=cavpl, \
                    cavrfact=cavrfact, theta_open=theta_open, zoffset=zoffset, \
                    t0=t0, tpl=tpl, dust=dust, aturb=aturb)
        else:
            self.envelope = TaperedUlrichEnvelopeExtended(mass=mass, rmin=rmin,\
                    rmax=rmax, rcent=rcent, gamma=gamma, cavpl=cavpl, \
                    cavrfact=cavrfact, theta_open=theta_open, zoffset=zoffset, \
                    t0=t0, tpl=tpl, dust=dust, aturb=aturb)

        if (dust != None):
            self.grid.add_density(self.envelope.density(self.grid.r, \
                    self.grid.theta, self.grid.phi),dust)

        if (gas != None):
            if (type(gas) == list):
                for i in range(len(gas)):
                    self.envelope.add_gas(gas[i], abundance[i])
                    self.grid.add_number_density(self.envelope.number_density( \
                            self.grid.r, self.grid.theta, self.grid.phi, \
                            gas=i), gas[i])
                    self.grid.add_velocity(self.envelope.velocity(self.grid.r, \
                            self.grid.theta, self.grid.phi, \
                            mstar=self.grid.stars[0].mass))
                    if aturb != None:
                        self.grid.add_microturbulence(self.envelope.\
                                microturbulence(self.grid.r, self.grid.theta, \
                                self.grid.phi))
            else:
                self.envelope.add_gas(gas, abundance)
                self.grid.add_number_density(self.envelope.number_density( \
                        self.grid.r, self.grid.theta, self.grid.phi, gas=0), \
                        gas)
                self.grid.add_velocity(self.envelope.velocity(self.grid.r, \
                        self.grid.theta, self.grid.phi, \
                        mstar=self.grid.stars[0].mass))
                if aturb != None:
                    self.grid.add_microturbulence(self.envelope.\
                            microturbulence(self.grid.r, self.grid.theta, \
                            self.grid.phi))

        if t0 != None:
            self.grid.add_temperature(self.envelope.temperature(self.grid.r, \
                    self.grid.theta, self.grid.phi))
        if tmid0 != None:
            self.grid.add_gas_temperature(self.envelope.gas_temperature( \
                    self.grid.r, self.grid.theta, self.grid.phi))

    def run_simple_dust_image(self, name=None, i=0., pa=0., npix=256, dx=1., \
            nu=230., kappa0=0.1, beta0=1, delta_beta=1, r0beta=100., plbeta=1, \
            dpc=140):
        # Convert inclination and position angle to radians.

        i *= numpy.pi / 180.
        pa *= numpy.pi / 180.

        # Set up the image plane.

        if npix%2 == 0:
            xx = numpy.linspace(-npix/2*dx*dpc, (npix/2-1)*dx*dpc, npix)
            yy = numpy.linspace(-npix/2*dx*dpc, (npix/2-1)*dx*dpc, npix)
        else:
            xx = numpy.linspace(-(npix-1)/2*dx*dpc, (npix-1)/2*dx*dpc, npix)
            yy = numpy.linspace(-(npix-1)/2*dx*dpc, (npix-1)/2*dx*dpc, npix)

        x, y = numpy.meshgrid(xx, yy)

        # Calculate physical coordinates from image plane coordinates.

        xpp = x*numpy.cos(pa) - y*numpy.sin(pa)
        ypp = (x*numpy.sin(pa) + y*numpy.cos(pa))/numpy.cos(i)
        rpp = numpy.sqrt(xpp**2 + ypp**2)
        phipp = numpy.arctan2(ypp, xpp) + pi/2

        # Calculate physical quantities.

        Sigma = self.disk.surface_density(rpp)

        T = self.disk.temperature_1d(rpp)

        B = B_nu(nu*1e9, T)

        beta = beta0 + delta_beta * (rpp / r0beta)**plbeta

        kappa = kappa0 * (nu / 1000.)**beta

        # Now do the actual calculation.

        I = B * (1. - numpy.exp(-kappa * Sigma / numpy.cos(i)))

        # Adjust the scale of I to be in the appropriate units.

        I = I / Jy * ((xx[1] - xx[0]) * AU / (dpc * pc)) * \
                ((yy[1] - yy[0]) * AU / (dpc * pc))

        self.images[name] =  Image(I.reshape((npix,npix,1,1)), x=xx/dpc, \
                y=yy/dpc, freq=numpy.array([nu])*1.0e9)

    def run_simple_dust_visibilities(self, name=None, i=0., pa=0., npix=256, \
            dx=1., nu=230., kappa0=0.1, beta0=1, delta_beta=1, r0beta=100., \
            plbeta=1, dpc=140):

        self.run_simple_dust_image(name="temp", i=i, pa=pa, npix=npix, \
                dx=dx, nu=nu, kappa0=kappa0, beta0=beta0, r0beta=r0beta, \
                delta_beta=delta_beta, plbeta=plbeta, dpc=dpc)

        self.visibilities[name] = imtovis(self.images["temp"])

        self.images.pop("temp")
    
    def run_simple_gas_image(self, i=0., pa=0., npix=256, dx=1., species=0, \
            trans=0, vstart=-10, dv=0.5, nv=40, n=0.5, T0=10000., plT=1, \
            v_z=0.):
        # Get a few constants from the molecule.

        A = self.disk.gas[species].A_ul[trans]
        nu0 = self.disk.gas[species].nu[trans]
        m_mol = self.disk.gas[species].mass * m_p

        # Set up the image plane.

        xx = numpy.linspace(-(npix-1)/2*dx, (npix-1)/2*dx, 256)
        yy = numpy.linspace(-(npix-1)/2*dx, (npix-1)/2*dx, 256)
        v = numpy.linspace(vstart*kms, (vstart+(nv-1)*dv)*kms, nv)
        nn = nu0 * (1 - v/c)

        x, y, nu = numpy.meshgrid(xx, yy, nn)

        # Calculate physical coordinates from image plane coordinates.

        xpp = -x*numpy.cos(pa) - y*numpy.sin(pa)
        ypp = (-x*numpy.sin(pa) + y*numpy.cos(pa))/numpy.cos(i)
        rpp = numpy.sqrt(xpp**2 + ypp**2)
        phipp = numpy.arctan2(ypp, xpp) + pi/2

        # Calculate physical quantities.

        Sigma = self.disk.surface_density(rpp)

        T = self.disk.temperature_1d(rpp, T_0=T0, p=plT)

        a_tot = numpy.sqrt(2*k*T/m_mol)
        print(a_tot.min(), a_tot.max())

        phi_dot_n = numpy.sin(phipp)*numpy.sin(i)*numpy.cos(pa) - \
                numpy.cos(phipp)*numpy.sin(i)*numpy.sin(pa)

        v_dot_n = numpy.sqrt(G*self.grid.stars[0].mass*M_sun/(rpp*AU)) * \
                phi_dot_n + v_z
        v_dot_n[(rpp >= self.disk.rmax) ^ (rpp <= self.disk.rmin)] = 0.0

        phi = numpy.zeros(rpp.shape)
        phi[a_tot != 0] = c / (a_tot[a_tot != 0]*nu0*numpy.sqrt(pi)) * \
                numpy.exp(-c**2*(nu[a_tot != 0]*(1 + v_dot_n[a_tot != 0]/c) - \
                nu0)**2 / (a_tot[a_tot != 0]**2 * nu0**2))

        # Now do the actual calculation.

        I = h*nu/(4*pi)*A*n*Sigma*phi

        return I, v
    
    def make_hyperion_symmetric(self):
        for i in range(len(self.grid.temperature)):
            ntheta = len(self.grid.theta)
            upper = self.grid.temperature[i][:,0:int(ntheta/2),:]
            lower = self.grid.temperature[i][:,int(ntheta/2):,:][:,::-1,:]
            average = 0.5 * (upper + lower)

            self.grid.temperature[i][:,0:int(ntheta/2),:] = average
            self.grid.temperature[i][:,int(ntheta/2):,:] =  average[:,::-1,:]

    def convert_hyperion_to_radmc3d(self):
        self.grid.r = self.grid.r[1:]
        self.grid.w1 = self.grid.w1[1:]

        ntheta = len(self.grid.theta)
        self.grid.theta = self.grid.theta[0:int(ntheta/2)]
        self.grid.w2 = self.grid.w2[0:int(ntheta/2+1)]
        
        for i in range(len(self.grid.density)):
            self.grid.density[i] = self.grid.density[i][1:,0:int(ntheta/2),:]
        for i in range(len(self.grid.temperature)):
            self.grid.temperature[i] = self.grid.temperature[i][1:,0:int(ntheta/2),:]
        for i in range(len(self.grid.number_density)):
            self.grid.number_density[i] = self.grid.number_density[i][1:,0:int(ntheta/2),:]
        for i in range(len(self.grid.microturbulence)):
            self.grid.microturbulence[i] = self.grid.microturbulence[i][1:,0:int(ntheta/2),:]
        for i in range(len(self.grid.velocity)):
            self.grid.velocity[i] = self.grid.velocity[i][:,1:,0:int(ntheta/2),:]

    def read_yso(self, filename):
        f = h5py.File(filename, "r")

        if ('Disk' in f):
            self.disk = Disk()
            self.disk.read(usefile=f['Disk'])

        if ('Envelope' in f):
            self.envelope = UlrichEnvelope()
            self.envelope.read(usefile=f['Envelope'])

        self.read(usefile=f)

        f.close()

    def write_yso(self, filename):
        f = h5py.File(filename, "w")

        self.write(usefile=f)

        if hasattr(self, "disk"):
            disk = f.create_group("Disk")
            self.disk.write(usefile=disk)

        if hasattr(self, "envelope"):
            envelope = f.create_group("Envelope")
            self.envelope.write(usefile=envelope)

        f.close()



import schwimmbad
import pickle
import numpy
import time
import os

def get_surrogate_model(params, model="pringle+ulrich+diana", \
        quantity="temperature", nthreads=1):

    # Load the keys for the parameters of the surrogate model.

    keys = list(numpy.loadtxt(os.path.dirname(os.path.abspath(__file__))+\
            "/surrogate_models/{0:s}/{1:s}/keys.txt".format(model, quantity), \
            dtype=str))

    # Load in the PCA that was found.

    pca = pickle.load(open(os.path.dirname(os.path.abspath(__file__))+\
            "/surrogate_models/{0:s}/{1:s}/pca.pkl".format(model, quantity), \
            "rb"))

    # Also the transformed data

    y_grid = numpy.load(os.path.dirname(os.path.abspath(__file__))+\
            "/surrogate_models/{0:s}/{1:s}/transformed_data.npy".\
            format(model, quantity))

    # Also load in the Gaussian process fits.

    gps = pickle.load(open(os.path.dirname(os.path.abspath(__file__))+\
            "/surrogate_models/{0:s}/{1:s}/gps.pkl".format(model, quantity), \
            "rb"))

    # Load the samples and make sure the best fit values are used for the 
    # hyperparameters.

    ncomponents = y_grid.shape[1]
    ncomponents = 9

    samples = []
    for i in range(ncomponents):
        samples += [numpy.load(os.path.dirname(os.path.abspath(__file__))+\
                "/surrogate_models/{0:s}/{1:s}/gp_samples_component{2:d}.pkl."
                "npy".format(model, quantity, i))]

    # Load in the data.

    x = []
    data = []

    x.append([(params[k] - -9.)**3 if k == "logM_env" else params[k] for k in \
            keys])

    x = numpy.array(x)
    shape = (99,100,1)

    # Set the parameter vector for the GPs randomly from the posteriors.

    for i in range(ncomponents):
        w = numpy.random.randint(samples[i].shape[0])

        gps[i].set_parameter_vector(samples[i][w])

    # Reconstruct the data from the PCA + GP fit.

    sample = lambda i: gps[i].sample_conditional(y_grid[:,i], x)[0] if \
            i < ncomponents else 0.

    t1 = time.time()
    with schwimmbad.JoblibPool(nthreads) as pool:
        components = [list(pool.map(sample, range(15)))]

    projected = pca.inverse_transform(components)
    t2 = time.time()
    print("Time to reconstruct = {0:f} seconds".format(t2 - t1))

    # Return the projected quantity.

    return 10.**projected[0].reshape(shape)



from ..constants.astronomy import arcsec
from ..constants.physics import c
from . import Image
try:
    import pvextractor
except:
    pass
import numpy

def extract_pv_diagram(image, xy=(0.,0.), pa=0., length=100, width=1):

    ny, nx, nfreq, npol = image.image.shape

    # Create a data cube that is appropriately shaped.

    cube = numpy.empty((nfreq, ny, nx))
    for i in range(nfreq):
        cube[i,:,:] = image.image[:,:,i,0]

    # Create the Path object.

    x0, y0 = xy

    line = [(x0-length/2*numpy.sin(pa),y0-length/2*numpy.cos(pa)), \
            (x0+length/2*numpy.sin(pa), y0+length/2*numpy.cos(pa))]

    path = pvextractor.Path(line, width=width)

    # Extract the PV diagram along the Path

    pv = pvextractor.extract_pv_slice(cube, path)

    # Convert back to my Image class.

    pvdiagram = numpy.empty((1, pv.data.shape[1], pv.data.shape[0], 1))
    for i in range(pv.data.shape[0]):
        pvdiagram[0,:,i,0] = pv.data[i,:]

    # Get the velocity.

    velocity = c * (image.header["RESTFRQ"] - image.freq) / \
            image.header["RESTFRQ"]

    # Get the x coordinates.

    x0 =  0.
    dx =  image.header["CDELT2"] * numpy.pi/180 / arcsec
    nx0 = int(pv.data.shape[1] / 2) + 1

    x = (numpy.arange(pv.data.shape[1]) - (nx0-1))*dx + x0

    return Image(pvdiagram, x=x, velocity=velocity, freq=image.freq)



import numpy
import astropy
import astropy.coordinates

#def update_catalog(catalog, *lists, tol=0.3, column_names=['1']):
def update_catalog(catalog, tol=0.3, column_names=['1'], *lists):

    coord_catalog = astropy.coordinates.SkyCoord(catalog["ra"].tolist(), \
            catalog["dec"].tolist(), frame='icrs')
    catalog["id"] = numpy.arange(len(catalog))

    for i, list1 in enumerate(lists):
        coord1 = astropy.coordinates.SkyCoord(list1["ra"].tolist(), \
                list1['dec'].tolist(), frame='icrs')

        idx, d2d, d3d = astropy.coordinates.match_coordinates_sky(coord1, \
                coord_catalog)

        ids, counts = numpy.unique(idx[d2d.arcsec < tol], return_counts=True)
        for j in range(len(ids)):
            if counts[j] > 1:
                min_d2d = min(d2d[idx == ids[j]].arcsec)
                d2d[(d2d.arcsec != min_d2d) & (idx == ids[j])] = \
                        astropy.coordinates.Angle("0d00m{0:f}s".format(2*tol))

        for j, k in enumerate(idx):
            if d2d.arcsec[j] < tol:
                for key in list1.colnames:
                    catalog[key+"_{0}".format(column_names[i])][k] = \
                            list1[key][j]

    del catalog["id"]

    return catalog



base_parameters = {
        # Stellar parameters.
        "logM_star":{"fixed":True, "value":0.0, "limits":[-1.,1.]},
        "T_star":{"fixed":True, "value":4000., "limits":[500.,10000.]},
        "logL_star":{"fixed":True, "value":0.0, "limits":[-1.,2.]},
        # Disk parameters.
        "disk_type":{"fixed":True, "value":"truncated", "limits":[0.,0.]},
        "logM_disk":{"fixed":True, "value":-4., "limits":[-10.,-2.5]},
        "logR_in":{"fixed":True, "value":-1., "limits":[-1.,4.]},
        "logR_disk":{"fixed":True, "value":2., "limits":[0.,4.]},
        "h_0":{"fixed":True, "value":0.1, "limits":[0.01,0.5]},
        "gamma":{"fixed":True, "value":1.0, "limits":[-0.5,2.0]},
        "gamma_taper":{"fixed":True, "value":"gamma", "limits":[-0.5,2.0]},
        "beta":{"fixed":True, "value":1.0, "limits":[0.5,1.5]},
        "logR_cav":{"fixed":True, "value":1.0, "limits":[-1.,3.]},
        "logdelta_cav":{"fixed":True, "value":0.0, "limits":[-4.,0.]},
        "logR_gap1":{"fixed":True, "value":1.0, "limits":[-1.,3.]},
        "w_gap1":{"fixed":True, "value":10., "limits":[1.,100.]},
        "logdelta_gap1":{"fixed":True, "value":0.0, "limits":[-4.,0.]},
        "logR_gap2":{"fixed":True, "value":1.4, "limits":[-1.,3.]},
        "w_gap2":{"fixed":True, "value":10., "limits":[1.,100.]},
        "logdelta_gap2":{"fixed":True, "value":0.0, "limits":[-4.,0.]},
        "logR_gap3":{"fixed":True, "value":1.8, "limits":[-1.,3.]},
        "w_gap3":{"fixed":True, "value":10., "limits":[1.,100.]},
        "logdelta_gap3":{"fixed":True, "value":0.0, "limits":[-4.,0.]},
        "f_M_large":{"fixed":True, "value":1., "limits":[0.05, 1.]},
        "logalpha_settle":{"fixed":True, "value":-2., "limits":[-5., 0.]},
        # Disk temperature parameters.
        "logT0":{"fixed":True, "value":2.5, "limits":[1.,3.]},
        "q":{"fixed":True, "value":0.25, "limits":[0.,1.]},
        "loga_turb":{"fixed":True, "value":-1.0, "limits":[-1.5,1.]},
        # Dartois temperature properties.
        "logTmid0":{"fixed":True, "value":2.0, "limits":[0.,3.]}, 
        "logTatm0":{"fixed":True, "value":2.5, "limits":[1.,3.]}, 
        "zq0":{"fixed":True, "value":0.1, "limits":[0.01,0.5]},
        "pltgas":{"fixed":True, "value":0.5, "limits":[0.,1.]},
        "delta":{"fixed":True, "value":1.0, "limits":[0.5,1.5]},
        # Envelope parameters.
        "envelope_type":{"fixed":True, "value":"none", "limits":[0.,0.]},
        "logM_env":{"fixed":True, "value":-3., "limits":[-10., -2.]},
        "logR_in_env":{"fixed":True, "value":"logR_in", "limits":[-1., 4.]},
        "logR_env":{"fixed":True, "value":3., "limits": [2.,5.]},
        "logR_c":{"fixed":True, "value":"logR_disk", "limits":[-1.,4.]},
        "f_cav":{"fixed":True, "value":0.5, "limits":[0.,1.]},
        "ksi":{"fixed":True, "value":1.0, "limits":[0.5,1.5]},
        "theta_open":{"fixed":True, "value":"45", "limits":[0.,90.]},
        "zoffset":{"fixed":True, "value":1, "limits":[0.,5.]},
        "gamma_env":{"fixed":True, "value":0., "limits":[-0.5,2.0]},
        # Envelope temperature parameters.
        "logT0_env":{"fixed":True, "value":2.5, "limits":[1.,3.5]},
        "q_env":{"fixed":True, "value":0.25, "limits":[0.,1.5]},
        "loga_turb_env":{"fixed":True, "value":-1.0, "limits":[-1.5,1.]},
        # Ambient medium?
        "ambient_medium":{"fixed":True, "value":False, "limits":[0.,0.]},
        # Dust parameters.
        "dust_file":{"fixed":True, "value":"pollack_new.hdf5", "limits":[0.,0.]},
        "loga_min":{"fixed":True, "value":-1.3, "limits":[0.,5.]},
        "loga_max":{"fixed":True, "value":0., "limits":[0.,5.]},
        "p":{"fixed":True, "value":3.5, "limits":[2.5,4.5]},
        "na":{"fixed":True, "value":100, "limits":[0,1000]},
        "envelope_dust":{"fixed":True, "value":"pollack_new.hdf5", "limits":[0.,0.]},
        # Gas parameters.
        "gas_file1":{"fixed":True, "value":"co.dat", "limits":[0.,0.]},
        "logabundance1":{"fixed":True, "value":-4., "limits":[-6.,-2.]},
        "freezeout1":{"fixed":True, "value":0., "limits":[0.,40.]},
        "mu":{"fixed":True, "value":2.37, "limits":[0.,0.]},
        # Viewing parameters.
        "i":{"fixed":True, "value":45., "limits":[0.,180.]},
        "pa":{"fixed":True, "value":0., "limits":[0.,360.]},
        "x0":{"fixed":True, "value":0., "limits":[-0.1,0.1]},
        "y0":{"fixed":True, "value":0., "limits":[-0.1,0.1]},
        "dpc":{"fixed":True, "value":140., "prior":"box", "sigma":0., "limits":[1.,1e6]},
        "Ak":{"fixed":True, "value":0., "limits":[0.,1.]},
        "v_sys":{"fixed":True, "value":5., "limits":[0.,10.]},
        "docontsub":{"fixed":True, "value":False, "limits":[0.,0.]},
        # Gas extinction parameters.
        "tau0":{"fixed":True, "value":0., "limits":[0.,10.]},
        "v_ext":{"fixed":True, "value":4., "limits":[2.,6.]},
        "sigma_vext":{"fixed":True, "value":1.0, "limits":[0.01,5.]},
        # Free-free emission.
        "logF_nu_ff":{"fixed":True, "value":-20., "limits":[-30.,0.]},
        "lognu_turn":{"fixed":True, "value":0., "limits":[-1.0,2.]},
        "pl_turn":{"fixed":True, "value":0.6, "limits":[0.,2.]},
        # Nuisance parameters.
        "flux_unc1":{"fixed":True, "value":1., "prior":"box", "sigma":0., "limits":[0.5,1.5]},
        "flux_unc2":{"fixed":True, "value":1., "prior":"box", "sigma":0., "limits":[0.5,1.5]},
        "flux_unc3":{"fixed":True, "value":1., "prior":"box", "sigma":0., "limits":[0.5,1.5]},
        }



#!/usr/bin/env python3

from pdspy.dust import *
import numpy

water_ice = Dust()
water_ice.set_optical_constants_from_henn("optical_constants/water_ice.txt")
water_ice.set_density(0.92)

silicates = Dust()
silicates.set_optical_constants_from_draine("optical_constants/astronomical_silicates.txt")
silicates.set_density(3.3)
silicates.calculate_optical_constants_on_wavelength_grid(water_ice.lam)

amorphous_carbon = Dust()
amorphous_carbon.set_optical_constants_from_henn("optical_constants/amorphous_carbon_zubko1996_extrapolated.txt")
#amorphous_carbon.set_density(2.24)
amorphous_carbon.set_density(1.0)
amorphous_carbon.calculate_optical_constants_on_wavelength_grid(water_ice.lam)

species = [silicates,amorphous_carbon,water_ice]
abundances = numpy.array([0.8,0.2,0.5])
abundances = abundances / abundances.sum()
print(abundances)

dust = mix_dust(species, abundances, filling=0.75)

amin = 0.05e-4
amax = 1.000e-4
pl = 3.5

dust.calculate_size_distribution_opacity(amin, amax, pl, with_dhs=True, \
        coat_volume_fraction=0.0, nang=1)

dust.write('diana_wice_1um.hdf5')



#!/usr/bin/env python3

from ..constants.astronomy import arcsec
from ..constants.physics import c
from .YSOModel import YSOModel
from .get_surrogate_model import get_surrogate_model
from .. import interferometry as uv
from .. import spectroscopy as sp
from .. import misc
from .. import dust
import scipy.signal
import subprocess
import tempfile
import signal
import numpy
import time
import os

import sys
if sys.version_info.major > 2:
    from subprocess import TimeoutExpired
else:
    from subprocess32 import TimeoutExpired

################################################################################
#
# Create a function which returns a model of the data.
#
################################################################################

def run_disk_model(visibilities, images, spectra, params, parameters, \
        plot=False, ncpus=1, ncpus_highmass=1, with_hyperion=False, \
        timelimit=3600, source="disk", nice=None, disk_vis=False, \
        no_radiative_transfer=False, nlam_SED=50, run_thermal=True, \
        surrogate=[], verbose=False, ftcode="galario", \
        percentile=99., absolute=2., relative=1.02, \
        increase_photons_until_convergence=False):

    # Set the values of all of the parameters.

    p = {}
    for key in parameters:
        if parameters[key]["fixed"]:
            if parameters[key]["value"] in parameters.keys():
                if parameters[parameters[key]["value"]]["fixed"]:
                    value = parameters[parameters[key]["value"]]["value"]
                else:
                    value = params[parameters[key]["value"]]
            else:
                value = parameters[key]["value"]
        else:
            value = params[key]

        p[key] = value
        if key[0:3] == "log":
            p[key[3:]] = 10.**value

    # Make sure alpha is defined.

    p["alpha"] = p["gamma"] + p["beta"]

    # If we're using a Pringle disk, make sure the scale height is set correctly

    if p["disk_type"] in ["exptaper","twolayerexptaper","settledexptaper"]:
        p["h_0"] *= p["R_disk"]**p["beta"]

    # Get the needed values of the gaps.

    p["R_in_gap1"] = p["R_gap1"] - p["w_gap1"]/2
    p["R_out_gap1"] = p["R_gap1"] + p["w_gap1"]/2
    p["R_in_gap2"] = p["R_gap2"] - p["w_gap2"]/2
    p["R_out_gap2"] = p["R_gap2"] + p["w_gap2"]/2
    p["R_in_gap3"] = p["R_gap3"] - p["w_gap3"]/2
    p["R_out_gap3"] = p["R_gap3"] + p["w_gap3"]/2

    # Set up the dust.

    dustopac = p["dust_file"]
    dust_gen = dust.DustGenerator(dust.__path__[0]+"/data/"+dustopac)
    if not p["disk_type"] in ["settled","settledexptaper"]:
        ddust = dust_gen(p["a_max"] / 1e4, p["p"])

    dustopac_env = p["envelope_dust"]
    env_dust_gen = dust.DustGenerator(dust.__path__[0]+"/data/"+dustopac_env)
    edust = env_dust_gen(1.0e-4, 3.5)

    # Make sure we are in a temp directory to not overwrite anything.

    original_dir = os.environ["PWD"]
    temp_dir = tempfile.TemporaryDirectory()
    os.chdir(temp_dir.name)

    # Write the parameters to a text file so it is easy to keep track of them.

    f = open("params.txt","w")
    for key in p:
        f.write("{0:s} = {1}\n".format(key, p[key]))
    f.close()

    # Set up the model and run the thermal simulation.

    if p["M_disk"] > 0.001 or p["R_disk"] < 50 or p["M_env"] > 0.001 or \
            p["R_env"] < 500 or p["h_0"] > 0.25:
        nprocesses = ncpus_highmass
    else:
        nprocesses = ncpus

    if with_hyperion:
        nphi = 201
        code = "hyperion"
    else:
        nphi = 101
        code = "radmc3d"

    m = YSOModel()
    m.add_star(mass=p["M_star"],luminosity=p["L_star"],temperature=p["T_star"])

    if p["envelope_type"] == "ulrich":
        p["R_grid"] = p["R_env"]
    else:
        #p["R_grid"] = max(5*p["R_disk"],300)
        p["R_grid"] = 5*p["R_disk"]
    m.set_spherical_grid(p["R_in"], p["R_grid"], 100, nphi, 2, code=code)

    if p["disk_type"] == "exptaper":
        m.add_pringle_disk(mass=p["M_disk"], rmin=p["R_in"], \
                rmax=p["R_disk"], plrho=p["alpha"], \
                h0=p["h_0"], plh=p["beta"], dust=ddust, \
                t0=p["T0"], plt=p["q"], \
                gap_rin=[p["R_in"],p["R_in_gap1"],p["R_in_gap2"],\
                p["R_in_gap3"]], gap_rout=[p["R_cav"],p["R_out_gap1"],\
                p["R_out_gap2"],p["R_out_gap3"]], gap_delta=[p["delta_cav"],\
                p["delta_gap1"],p["delta_gap2"],p["delta_gap3"]], \
                gamma_taper=p["gamma_taper"])
    elif p["disk_type"] == "twolayer":
        m.add_twolayer_disk(mass=p["M_disk"], rmin=p["R_in"], rmax=p["R_disk"],\
                plrho=p["alpha"], h0=p["h_0"], plh=p["beta"], dust=dust_gen,\
                gap_rin=[p["R_in"],p["R_in_gap1"],p["R_in_gap2"],\
                p["R_in_gap3"]], gap_rout=[p["R_cav"],p["R_out_gap1"],\
                p["R_out_gap2"],p["R_out_gap3"]], gap_delta=[p["delta_cav"],\
                p["delta_gap1"],p["delta_gap2"],p["delta_gap3"]], \
                amin=p["a_min"], amax=p["a_max"], fmin=p["f_M_large"], \
                alpha_settle=p["alpha_settle"])
    elif p["disk_type"] == "twolayerexptaper":
        m.add_twolayer_pringle_disk(mass=p["M_disk"], rmin=p["R_in"], \
                rmax=p["R_disk"],plrho=p["alpha"], h0=p["h_0"], plh=p["beta"], \
                dust=dust_gen,gap_rin=[p["R_in"],p["R_in_gap1"],p["R_in_gap2"],\
                p["R_in_gap3"]], gap_rout=[p["R_cav"],p["R_out_gap1"],\
                p["R_out_gap2"],p["R_out_gap3"]], gap_delta=[p["delta_cav"],\
                p["delta_gap1"],p["delta_gap2"],p["delta_gap3"]], \
                amin=p["a_min"], amax=p["a_max"], fmin=p["f_M_large"], \
                alpha_settle=p["alpha_settle"])
    elif p["disk_type"] == "settled":
        m.add_settled_disk(mass=p["M_disk"], rmin=p["R_in"], rmax=p["R_disk"], \
                plrho=p["alpha"], h0=p["h_0"], plh=p["beta"], dust=dust_gen,\
                gap_rin=[p["R_in"],p["R_in_gap1"],p["R_in_gap2"],\
                p["R_in_gap3"]], gap_rout=[p["R_cav"],p["R_out_gap1"],\
                p["R_out_gap2"],p["R_out_gap3"]], gap_delta=[p["delta_cav"],\
                p["delta_gap1"],p["delta_gap2"],p["delta_gap3"]], \
                amin=p["a_min"], amax=p["a_max"], pla=p["p"], na=p["na"], \
                alpha_settle=p["alpha_settle"])
    elif p["disk_type"] == "settledexptaper":
        m.add_settled_pringle_disk(mass=p["M_disk"], rmin=p["R_in"], \
                rmax=p["R_disk"], plrho=p["alpha"], h0=p["h_0"], plh=p["beta"],\
                dust=dust_gen,gap_rin=[p["R_in"],p["R_in_gap1"],p["R_in_gap2"],\
                p["R_in_gap3"]], gap_rout=[p["R_cav"],p["R_out_gap1"],\
                p["R_out_gap2"],p["R_out_gap3"]], gap_delta=[p["delta_cav"],\
                p["delta_gap1"],p["delta_gap2"],p["delta_gap3"]], \
                amin=p["a_min"], amax=p["a_max"], pla=p["p"], na=p["na"], \
                alpha_settle=p["alpha_settle"], gamma_taper=p["gamma_taper"])
    else:
        m.add_disk(mass=p["M_disk"], rmin=p["R_in"], \
                rmax=p["R_disk"], plrho=p["alpha"], \
                h0=p["h_0"], plh=p["beta"], dust=ddust, \
                t0=p["T0"], plt=p["q"], \
                gap_rin=[p["R_in"],p["R_in_gap1"],p["R_in_gap2"],\
                p["R_in_gap3"]], gap_rout=[p["R_cav"],p["R_out_gap1"],\
                p["R_out_gap2"],p["R_out_gap3"]], gap_delta=[p["delta_cav"],\
                p["delta_gap1"],p["delta_gap2"],p["delta_gap3"]])

    if p["envelope_type"] == "ulrich":
        m.add_ulrich_envelope(mass=p["M_env"], rmin=p["R_in"], rmax=p["R_env"],\
                cavpl=p["ksi"], cavrfact=p["f_cav"], t0=p["T0_env"], \
                tpl=p["q_env"], aturb=p["a_turb_env"], dust=edust)
    elif p["envelope_type"] == "ulrich-extended":
        m.add_ulrichextended_envelope(mass=p["M_env"], rmin=p["R_in"], \
                rmax=p["R_env"], cavpl=p["ksi"], cavrfact=p["f_cav"], 
                theta_open=p["theta_open"], zoffset=p["zoffset"], \
                t0=p["T0_env"], tpl=p["q_env"], aturb=p["a_turb_env"], \
                dust=edust)
    elif p["envelope_type"] == "ulrich-tapered":
        m.add_tapered_ulrich_envelope(mass=p["M_env"], rmin=p["R_in"], \
                rmax=p["R_env"], gamma=p["gamma_env"], cavpl=p["ksi"], \
                cavrfact=p["f_cav"], dust=edust, t0=p["T0_env"], \
                tpl=p["q_env"], aturb=p["a_turb_env"], rcent=p["R_c"])
    elif p["envelope_type"] == "ulrich-tapered-extended":
        m.add_tapered_ulrichextended_envelope(mass=p["M_env"], rmin=p["R_in"], \
                rmax=p["R_env"], gamma=p["gamma_env"], rcent=p["R_c"], \
                cavpl=p["ksi"], cavrfact=p["f_cav"], \
                theta_open=p["theta_open"], zoffset=p["zoffset"], \
                t0=p["T0_env"], tpl=p["q_env"], aturb=p["a_turb_env"], \
                dust=edust)
    else:
        pass

    if p["ambient_medium"]:
        m.add_ambient_medium()

    m.grid.set_wavelength_grid(0.1,1.0e5,500,log=True)

    # If we just want to generate the model, quit here.

    if no_radiative_transfer:
        os.system("rm params.txt")
        os.chdir(original_dir)

        return m

    # Run the thermal simulation.

    if "temperature" in surrogate:
        # Use the surrogate model to calculate the temperature.

        temperature = get_surrogate_model(p, quantity="temperature")

        m.grid.temperature = []
        for i in range(len(m.grid.density)):
            m.grid.add_temperature(temperature)
    else:
        if code == "hyperion" and run_thermal:
            try:
                m.run_thermal(code="hyperion", nphot=2e5, mrw=True, pda=True, \
                        niterations=10, mpi=True if nprocesses > 1 else False, \
                        nprocesses=nprocesses, verbose=verbose, \
                        timeout=timelimit, percentile=percentile, \
                        absolute=absolute, relative=relative, \
                        increase_photons_until_convergence=\
                        increase_photons_until_convergence)

                # Convert model to radmc-3d format.

                m.make_hyperion_symmetric()

                m.convert_hyperion_to_radmc3d()
            except:
                os.chdir(original_dir)
                return 0.
        elif run_thermal:
            try:
                t1 = time.time()
                m.run_thermal(code="radmc3d", nphot=1e6, \
                        modified_random_walk=True,\
                        mrw_gamma=2, mrw_tauthres=10, mrw_count_trigger=100, \
                        verbose=verbose, setthreads=nprocesses, \
                        timelimit=timelimit, nice=nice)
                t2 = time.time()
                f = open(original_dir + "/times.txt", "a")
                f.write("{0:f}\n".format(t2-t1))
                f.close()
            # Catch a timeout error from models running for too long...
            except TimeoutExpired:
                t2 = time.time()
                f = open(original_dir + "/times.txt", "a")
                f.write("{0:f}\n".format(t2-t1))
                f.close()

                os.system("mv params.txt {0:s}/params_timeout_{1:s}".format(\
                        original_dir, time.strftime("%Y-%m-%d-%H:%M:%S", \
                        time.gmtime())))
                os.system("rm *.inp *.out *.dat *.uinp")
                os.chdir(original_dir)

                return 0.
            # Catch a strange RADMC3D error and re-run. Not an ideal fix, but 
            # perhaps the simplest option.
            except:
                t1 = time.time()
                m.run_thermal(code="radmc3d", nphot=1e6, \
                        modified_random_walk=True,\
                        mrw_gamma=2, mrw_tauthres=10, mrw_count_trigger=100, \
                        verbose=verbose, setthreads=nprocesses, \
                        timelimit=timelimit, nice=nice)
                t2 = time.time()
                f = open(original_dir + "/times.txt", "a")
                f.write("{0:f}\n".format(t2-t1))
                f.close()

    # Run the images/visibilities/SEDs. If plot == "concat" then we are doing
    # a fit and we need less. Otherwise we are making a plot of the best fit 
    # model so we need to generate a few extra things.

    # Run the visibilities.

    for j in range(len(visibilities["file"])):
        # Set the wavelengths for RADMC3D to use.

        wave = c / visibilities["data"][j].freq / 1.0e-4
        m.set_camera_wavelength(wave)

        if ftcode == "galario":
            m.run_image(name=visibilities["lam"][j], nphot=1e5, \
                    npix=visibilities["npix"][j], \
                    pixelsize=visibilities["pixelsize"][j], \
                    lam=None, loadlambda=True, incl=p["i"], \
                    pa=p["pa"], dpc=p["dpc"], code="radmc3d", \
                    mc_scat_maxtauabs=5, verbose=verbose,setthreads=nprocesses,\
                    writeimage_unformatted=True, nice=nice)
        else:
            m.run_image(name=visibilities["lam"][j], nphot=1e5, \
                    lam=None, loadlambda=True, incl=p["i"], \
                    pa=p["pa"], dpc=p["dpc"], code="radmc3d", \
                    mc_scat_maxtauabs=5, verbose=verbose,setthreads=nprocesses,\
                    writeimage_unformatted=True, nice=nice, unstructured=True)

        # Account for the flux calibration uncertainties.

        m.images[visibilities["lam"][j]].image *= p["flux_unc{0:d}".format(j+1)]

        m.visibilities[visibilities["lam"][j]] = uv.interpolate_model(\
                visibilities["data"][j].u, visibilities["data"][j].v, \
                visibilities["data"][j].freq, m.images[visibilities["lam"][j]],\
                dRA=p["x0"], dDec=p["y0"], nthreads=nprocesses, code=ftcode, \
                nxy=visibilities["npix"][j], dxy=visibilities["pixelsize"][j])

        # Add in free free emission.

        m.visibilities[visibilities["lam"][j]].real += uv.model(\
                m.visibilities[visibilities["lam"][j]].u, \
                m.visibilities[visibilities["lam"][j]].v, \
                [p["x0"],p["y0"],sp.freefree(m.visibilities[visibilities[\
                "lam"][j]].freq.mean(), p["F_nu_ff"], p["nu_turn"]*1e9, \
                p["pl_turn"])], return_type="data", funct="point").real

        if plot:
            # Make high resolution visibilities. 

            if "galario" in ftcode:
                u, v = numpy.meshgrid(numpy.linspace(-2.0e6, 2.0e6, 2000), \
                        numpy.linspace(-2.0e6, 2.0e6, 2000))
            else:
                u, v = numpy.meshgrid(numpy.hstack((\
                        -numpy.logspace(3.,7.,50)[::-1],\
                        numpy.logspace(3.,7.,50))),\
                        numpy.hstack((-numpy.logspace(3.,7.,50)[::-1],\
                        numpy.logspace(3.,7.,50))))

            u, v = u.reshape((u.size,)), v.reshape((v.size,))

            m.visibilities[visibilities["lam"][j]+"_high"] = \
                    uv.interpolate_model(u, v, visibilities["data"][j].freq, \
                    m.images[visibilities["lam"][j]], dRA=p["x0"], \
                    dDec=p["y0"], nthreads=nprocesses, code=ftcode, \
                    nxy=visibilities["npix"][j], \
                    dxy=visibilities["pixelsize"][j])

            # Add in free free emission.

            m.visibilities[visibilities["lam"][j]+"_high"].real += uv.model(\
                    m.visibilities[visibilities["lam"][j]+"_high"].u, \
                    m.visibilities[visibilities["lam"][j]+"_high"].v, \
                    [p["x0"],p["y0"],sp.freefree(m.visibilities[visibilities[\
                    "lam"][j]].freq.mean(), p["F_nu_ff"], p["nu_turn"]*1e9, \
                    p["pl_turn"])], return_type="data", funct="point").real

            # Run the 2D visibilities.

            m.visibilities[visibilities["lam"][j]+"_2d"] = \
                    uv.interpolate_model(visibilities["data2d"][j].u, \
                    visibilities["data2d"][j].v, \
                    visibilities["data2d"][j].freq, \
                    m.images[visibilities["lam"][j]], dRA=p["x0"], \
                    dDec=p["y0"], nthreads=nprocesses, code=ftcode)

            # Run a millimeter image.

            wave = c / visibilities["image"][j].freq / 1.0e-4
            m.set_camera_wavelength(wave)

            m.run_image(name=visibilities["lam"][j], nphot=1e5, \
                    npix=visibilities["image_npix"][j], \
                    pixelsize=visibilities["image_pixelsize"][j], \
                    lam=None, loadlambda=True, incl=p["i"], \
                    pa=p["pa"], dpc=p["dpc"], code="radmc3d", \
                    mc_scat_maxtauabs=5, verbose=verbose, \
                    setthreads=nprocesses, nice=nice)

            m.images[visibilities["lam"][j]].image[\
                    int(visibilities["image_npix"][j]/2),
                    int(visibilities["image_npix"][j]/2),0,0] += \
                    sp.freefree(m.images[visibilities["lam"][j]].freq.mean(), \
                    p["F_nu_ff"],p["nu_turn"]*1e9,p["pl_turn"])

            x, y = numpy.meshgrid(numpy.linspace(-256,255,512), \
                    numpy.linspace(-256,255,512))

            beam = misc.gaussian2d(x, y, 0., 0., \
                    visibilities["image"][j].header["BMAJ"]/2.355/\
                    visibilities["image"][j].header["CDELT2"], \
                    visibilities["image"][j].header["BMIN"]/2.355/\
                    visibilities["image"][j].header["CDELT2"], \
                    (90-visibilities["image"][j].header["BPA"])*numpy.pi/180., \
                    1.0)

            m.images[visibilities["lam"][j]].image = scipy.signal.fftconvolve(\
                    m.images[visibilities["lam"][j]].image[:,:,0,0], beam, \
                    mode="same").reshape(m.images[visibilities["lam"][j]].\
                    image.shape)

            # Run visibilities that include only the contribution of the disk.

            if disk_vis and parameters["envelope_type"]["value"] == "ulrich":
                density_original = m.grid.density.copy()
                temperature_original = m.grid.temperature.copy()
                dust_original = m.grid.dust.copy()

                del m.grid.density[-1]
                del m.grid.temperature[-1]
                del m.grid.dust[-1]

                wave = c / visibilities["data"][j].freq / 1.0e-4
                m.set_camera_wavelength(wave)

                if ftcode == "galario":
                    m.run_image(name=visibilities["lam"][j]+"_disk", nphot=1e5,\
                            npix=visibilities["npix"][j], \
                            pixelsize=visibilities["pixelsize"][j], \
                            lam=None, loadlambda=True, incl=p["i"], \
                            pa=p["pa"], dpc=p["dpc"], code="radmc3d", \
                            mc_scat_maxtauabs=5, verbose=verbose, \
                            setthreads=nprocesses,writeimage_unformatted=True, \
                            nice=nice)
                else:
                    m.run_image(name=visibilities["lam"][j]+"_disk", nphot=1e5,\
                            lam=None, loadlambda=True, incl=p["i"], \
                            pa=p["pa"], dpc=p["dpc"], code="radmc3d", \
                            mc_scat_maxtauabs=5, verbose=verbose, \
                            setthreads=nprocesses, writeimage_unformatted=True,\
                            nice=nice, unstructured=True)

                m.visibilities[visibilities["lam"][j]+"_disk"] = \
                        uv.interpolate_model(u,v,visibilities["data"][j].freq, \
                        m.images[visibilities["lam"][j]+"_disk"], dRA=p["x0"], \
                        dDec=p["y0"], nthreads=nprocesses, code=ftcode, \
                        nxy=visibilities["npix"][j], \
                        dxy=visibilities["pixelsize"][j])

                m.grid.density = density_original
                m.grid.temperature = temperature_original
                m.grid.dust = dust_original

    # Run the images.

    for j in range(len(images["file"])):
        m.run_image(name=images["lam"][j], nphot=1e5, \
                npix=images["npix"][j], pixelsize=images["pixelsize"][j], \
                lam=images["lam"][j], incl=p["i"], \
                pa=p["pa"], dpc=p["dpc"], code="radmc3d", \
                mc_scat_maxtauabs=5, verbose=verbose, setthreads=nprocesses, \
                nice=nice)

        # Convolve with the beam.

        x, y = numpy.meshgrid(numpy.linspace(-256,255,512), \
                numpy.linspace(-256,255,512))

        beam = misc.gaussian2d(x, y, 0., 0., images["bmaj"][j]/2.355/\
                images["pixelsize"][j], images["bmin"][j]/2.355/\
                image["pixelsize"][j], (90-images["bpa"][j])*numpy.pi/180., 1.0)

        m.images[images["lam"][j]].image = scipy.signal.fftconvolve(\
                m.images[images["lam"][j]].image[:,:,0,0], beam, mode="same").\
                reshape(m.images[images["lam"][j]].image.shape)

    # Run the SED.

    if "total" in spectra:
        if plot:
            m.set_camera_wavelength(numpy.logspace(-1,4,nlam_SED))
        else:
            m.set_camera_wavelength(spectra["total"].wave)

        m.run_sed(name="SED", nphot=1e4, loadlambda=True, incl=p["i"],\
                pa=p["pa"], dpc=p["dpc"], code="radmc3d", \
                camera_scatsrc_allfreq=True, mc_scat_maxtauabs=5, \
                verbose=verbose, setthreads=nprocesses, nice=nice)

        # Add in a contribution from free-free emission.

        m.spectra["SED"].flux += sp.freefree(m.spectra["SED"].freq, \
                p["F_nu_ff"], p["nu_turn"]*1e9, p["pl_turn"])

        # Redden the SED based on the reddening.

        m.spectra["SED"].flux = dust.redden(m.spectra["SED"].wave, \
                m.spectra["SED"].flux, p["Ak"], law="mcclure")

        # Now take the log of the SED.

        if not plot:
            m.spectra["SED"].flux = numpy.log10(m.spectra["SED"].flux)

    # Clean up everything and return.

    os.system("rm params.txt")
    os.chdir(original_dir)

    return m



import numpy
from ..interferometry import Visibilities
from ..constants.astronomy import pc, arcsec
from scipy.fftpack import fft2, fftshift, fftfreq, ifftshift

def imtovis(image):

    ##### Some natural constants
    
    real = numpy.empty((image.x.size*image.y.size,image.freq.size))
    imag = numpy.empty((image.x.size*image.y.size,image.freq.size))
    weights = numpy.ones(real.shape)
    for i in range(image.freq.size):
        vis = fftshift(fft2(ifftshift(image.image[:,:,i,0])))
        real[:,i] = vis.real.reshape((image.x.size*image.y.size,))
        imag[:,i] = vis.imag.reshape((image.x.size*image.y.size,))

    uu = fftshift(fftfreq(image.x.size, (image.x[1] - image.x[0]) * arcsec))
    vv = fftshift(fftfreq(image.y.size, (image.y[1] - image.y[0]) * arcsec))
    
    u, v = numpy.meshgrid(uu, vv)
    u = u.reshape((image.x.size*image.y.size,))
    v = v.reshape((image.y.size*image.y.size,))

    freq = image.freq

    return Visibilities(u, v, freq, real, imag, weights)
